import datetime
import os

from django.conf import settings
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.db import models
from django.db.models import Avg, F, ImageField
from django.urls import reverse
from django.utils import timezone
from django.utils.safestring import mark_safe
from django.views.decorators.debug import sensitive_variables

from ddm.auth.models import ProjectAccessToken
from ddm.core.utils.misc import create_asciidigits_id
from ddm.datadonation.models import DataDonation
from ddm.encryption.models import Encryption
from ddm.logging.models import ExceptionLogEntry, ExceptionRaisers, EventLogEntry
from ddm.participation.models import Participant


class ResearchProfile(models.Model):
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name='ddm_research_profile',
    )
    active = models.BooleanField(default=True)
    created = models.DateTimeField('date registered', default=timezone.now)
    ignore_email_restriction = models.BooleanField(default=False)

    def __str__(self):
        return self.user.email


def project_header_dir_path(instance, filename):
    return f'project_{instance.pk}/headers/{filename}'


class DonationProject(models.Model):
    url_id = models.SlugField(unique=True, max_length=8)

    # Basic information for internal organization.
    name = models.CharField(
        max_length=50,
        help_text=(
            'Project Name - for internal organisation only (can still be '
            'changed later).'
        )
    )

    # Note: Value of "date_created" attribute must not be changed after
    # instance creation as it is used for en-/decryption.
    date_created = models.DateTimeField(default=timezone.now)
    owner = models.ForeignKey(
        'ResearchProfile',
        related_name='project_owner',
        on_delete=models.SET_NULL,
        null=True
    )

    # Public information.
    contact_information = models.TextField(
        blank=False,
        verbose_name='Contact information',
        help_text=(
            'Please provide the contact information of the person responsible '
            'for the conduction of this study (Name, professional address, e-mail, '
            'tel, ...). The contact information will be accessible for '
            'participants during the data donation. The field is mandatory.'
        )
    )
    data_protection_statement = models.TextField(
        blank=False,
        verbose_name='Data Protection Statement',
        help_text=(
            'Please provide a data protection statement for your data donation '
            'collection. This should include the purpose for which the data is '
            'collected, how it will be stored, and who will have access to the '
            'data. The field is mandatory.'
        ),
    )

    # Information affecting participation flow.
    slug = models.SlugField(
        unique=True,
        verbose_name='URL Identifier',
        help_text=mark_safe(
            'Identifier that is included in the URL through which participants '
            'can access the project '
            '(e.g, https://root.url/<b>my-url-identifier</b>). '
            'Can only contain letters, hyphens, numbers or underscores.'
        )
    )
    briefing_text = models.TextField(
        blank=False,
        verbose_name='Briefing Text',
        help_text=(
            'This text will be displayed to participants in the first step before '
            'beginning the data donation. Here, you should introduce your study, '
            'who is responsible for the study, provide a quick summary of what '
            'your participants will be asked to do in the next steps, etc.'
        )
    )
    briefing_consent_enabled = models.BooleanField(
        default=False,
        verbose_name='Briefing Consent Mandatory',
        help_text='Enable this option to obtain explicit consent from participant, '
                  'that they want to start the study.'
    )
    briefing_consent_label_yes = models.CharField(max_length=255, blank=True)
    briefing_consent_label_no = models.CharField(max_length=255, blank=True)
    debriefing_text = models.TextField(
        blank=False,
        verbose_name='Debriefing Text',
        help_text='This text will be displayed to participants on the last page '
                  'of the study (i.e., after the data donation and the questionnaire).'
    )

    # Appearance settings.
    img_header_left = ImageField(
        upload_to=project_header_dir_path,
        null=True,
        blank=True,
        verbose_name='Header Image Left'
    )
    img_header_right = ImageField(
        upload_to=project_header_dir_path,
        null=True,
        blank=True,
        verbose_name='Header Image Right'
    )

    # Access settings.
    public_key = models.BinaryField()
    super_secret = models.BooleanField(default=False)

    # Redirect settings.
    redirect_enabled = models.BooleanField(default=False, verbose_name='Redirect enabled')
    redirect_target = models.CharField(
        max_length=2000,
        blank=True,
        verbose_name='Redirect address',
        help_text=mark_safe(
            'Always include <i>http://</i> or <i>https://</i> in the redirect address. '
            'If URL parameter extraction is enabled for this project, you can '
            'include the extracted URL parameters in the redirect address as follows: '
            '"https://redirect.me/?redirectpara=<b>{{participant.data.url_param.URLParameter}}</b>".')
    )

    # URL parameter extraction settings.
    url_parameter_enabled = models.BooleanField(default=False, verbose_name='URL parameter extraction enabled')
    expected_url_parameters = models.CharField(
        max_length=500,
        blank=True,
        verbose_name='Expected URL parameter',
        help_text='Separate multiple parameters with a semikolon (";"). '
                  'Semikolons are not allowed as part of the expected url parameters.'
    )

    active = models.BooleanField(
        default=True,
        verbose_name='Active',
        help_text='Participants can only take part in a project if it is active.'
    )

    @sensitive_variables()
    def __init__(self, *args, **kwargs):
        self.secret_key = settings.SECRET_KEY
        if 'secret_key' in kwargs:
            self.secret_key = kwargs['secret_key']
            kwargs.pop('secret_key')
        super().__init__(*args, **kwargs)

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse('ddm_projects:detail', args=[str(self.url_id)])

    def get_salt(self):
        return str(self.date_created)

    def clean_file_on_reupload(self):
        """
        Deletes previous file on model FileFields or ImageFields if a file
        is re-uploaded.
        """
        i = DonationProject.objects.get(pk=self.pk)
        for field in i._meta.fields:
            if field.get_internal_type() in ['FileField', 'ImageField']:
                if getattr(i, field.name) != getattr(self, field.name):
                    file_attr = getattr(i, field.name)
                    try:
                        if os.path.isfile(file_attr.path):
                            os.remove(file_attr.path)
                    except ValueError:
                        pass
        return

    @sensitive_variables()
    def save(self, *args, **kwargs):
        if not self.pk:
            if self.owner is None:
                raise ValidationError(
                    'DonationProject.owner cannot be null on create.')
        else:
            self.clean_file_on_reupload()

        if not self.public_key:
            self.public_key = Encryption.get_public_key(self.secret, str(self.date_created))

        if not self.url_id:
            self.url_id = create_asciidigits_id(8)
            while DonationProject.objects.filter(url_id=self.url_id).exists():
                self.url_id = create_asciidigits_id(8)

        super().save(*args, **kwargs)

    @property
    def secret(self):
        return self.secret_key

    @secret.setter
    def secret(self, value):
        self.secret_key = value

    def delete(self, using=None, keep_parents=False):
        # Manually delete related questions to circumvent foreign key integrity
        # problem caused by QuestionBase models inheriting from django-polymorphic.
        # (see https://github.com/django-polymorphic/django-polymorphic/issues/34#issuecomment-1027866872)
        for question in self.questionbase_set.all():
            question.delete()
        return super().delete(using, keep_parents)

    def get_statistics(self):
        participants = Participant.objects.filter(project=self)
        statistics = {
            'n_started': participants.count(),
            'n_completed': participants.filter(completed=True).count(),
            'completion_rate': self.get_completion_rate(participants),
            'n_donations': DataDonation.objects.filter(project=self, status='success').count(),
            'n_errors': ExceptionLogEntry.objects.filter(project=self).count(),
            'average_time': self.get_average_completion_time(participants),
        }
        return statistics

    def get_completion_rate(self, participants=None):
        if participants is None:
            participants = Participant.objects.filter(project=self)
        if participants.count() > 0:
            return participants.filter(completed=True).count() / participants.count()
        else:
            return 0

    def get_average_completion_time(self, participants=None):
        if participants is None:
            participants = Participant.objects.filter(project=self)
        return str(participants.filter(completed=True).aggregate(
            v=Avg(F('end_time') - F('start_time')))['v']).split(".")[0]

    def get_questionnaire_config(self, participant, view):
        """
        Returns a dictionary containing all information to render the
        questionnaire for a given participant.
        """
        q_config = []
        questions = self.questionbase_set.all().order_by('page', 'index')
        for question in questions:
            if question.is_general():
                q_config.append(question.get_config(participant, view))
            else:
                try:
                    donation = DataDonation.objects.get(
                        blueprint=question.blueprint,
                        participant=participant
                    )
                except ObjectDoesNotExist:
                    msg = ('Questionnaire Rendering Exception: No donation '
                           f'found for participant {participant.pk} and '
                           f'blueprint {question.blueprint.pk}.')
                    ExceptionLogEntry.objects.create(
                        project=self,
                        raised_by=ExceptionRaisers.SERVER,
                        message=msg
                    )
                    continue

                if donation.consent and donation.status == 'success':
                    q_config.append(question.get_config(participant, view))
        return q_config

    def get_expected_url_parameters(self):
        return self.expected_url_parameters.split(';')

    def get_token(self):
        return ProjectAccessToken.objects.filter(project=self).first()

    def create_token(self, expiration_days=None):
        token = self.get_token()
        if token:
            token.delete()
        if expiration_days:
            expiration_date = timezone.now() + datetime.timedelta(days=expiration_days)
        else:
            expiration_date = None
        EventLogEntry.objects.create(
            project=self, description='New Access Token Created')
        return ProjectAccessToken.objects.create(
            project=self, expiration_date=expiration_date)



from django.contrib.messages.views import SuccessMessageMixin
from django.db.models import Max
from django.forms import inlineformset_factory
from django.urls import reverse
from django.views.generic import ListView
from django.views.generic.edit import CreateView, UpdateView, DeleteView

from django_ckeditor_5.widgets import CKEditor5Widget

from ddm.auth.views import DDMAuthMixin
from ddm.datadonation.models import DonationBlueprint
from ddm.projects.models import DonationProject
from ddm.questionnaire.models import (
    QuestionBase, QuestionType, SingleChoiceQuestion, MultiChoiceQuestion,
    OpenQuestion, MatrixQuestion, SemanticDifferential, Transition,
    QuestionItem, ScalePoint
)


class ProjectMixin:
    """ Mixin for all blueprint related views. """
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({'project_url_id': self.kwargs['project_url_id']})
        context.update({'project': self.get_project()})
        return context

    def get_project(self):
        return DonationProject.objects.get(url_id=self.kwargs['project_url_id'])


class QuestionnaireOverview(ProjectMixin, DDMAuthMixin, ListView):
    """ View to list all donation blueprints associated with a project. """
    model = DonationBlueprint
    context_object_name = 'donation_blueprints'
    template_name = 'ddm_questionnaire/list.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        question_types = QuestionType.choices
        context.update({
            'question_types': question_types,
            'questions': self.get_all_questions()
        })
        return context

    def get_all_questions(self):
        project = self.get_project()
        return project.questionbase_set.all()

    def get_queryset(self):
        return super().get_queryset().filter(
            project__url_id=self.kwargs['project_url_id'])


class QuestionFormMixin(ProjectMixin):
    model = None
    fields = '__all__'

    QUESTION_CLASSES = {
        'single_choice': SingleChoiceQuestion,
        'multi_choice': MultiChoiceQuestion,
        'open': OpenQuestion,
        'matrix': MatrixQuestion,
        'semantic_diff': SemanticDifferential,
        'transition': Transition
    }

    SHARED_FIELDS = [
        'name',
        'blueprint',
        'page',
        'index',
        'variable_name',
        'text',
        'required'
    ]
    QUESTION_FIELDS = {
        'single_choice': SHARED_FIELDS + ['randomize_items'],
        'multi_choice': SHARED_FIELDS + ['randomize_items'],
        'matrix': SHARED_FIELDS + ['randomize_items'],
        'semantic_diff': SHARED_FIELDS + ['randomize_items'],
        'open': SHARED_FIELDS + ['input_type', 'max_input_length', 'display'],
        'transition': SHARED_FIELDS
    }

    def get_queryset(self):
        self.model = self.QUESTION_CLASSES[self.kwargs['question_type']]
        self.fields = self.QUESTION_FIELDS[self.kwargs['question_type']]
        return super().get_queryset()

    def get_form(self, form_class=None):
        form = super().get_form(form_class)
        if self.kwargs['question_type'] == 'transition':
            del form.fields['required']
        return form

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        question_type_label = QuestionType(self.kwargs['question_type']).label
        context.update({'question_type': question_type_label})
        context['form'].fields['blueprint'].queryset = DonationBlueprint.objects.filter(
            project__url_id=self.kwargs['project_url_id'])
        context['form'].fields['blueprint'].empty_label = 'General Question – no Blueprint assigned'
        context['form'].fields['text'].widget = CKEditor5Widget(config_name='ddm_ckeditor_temp_func')
        return context


class QuestionCreate(SuccessMessageMixin, DDMAuthMixin, QuestionFormMixin, CreateView):
    """ View to create question. """
    template_name = 'ddm_questionnaire/create.html'
    success_message = 'New %(question_type)s was created.'

    def get_initial(self):
        initial = super().get_initial()
        initial['question_type'] = self.kwargs['question_type']
        return initial

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['instance'] = self.QUESTION_CLASSES[self.kwargs['question_type']](
            project=self.get_project(),
        )
        return kwargs

    def get_success_url(self):
        kwargs = {
            'project_url_id': self.kwargs['project_url_id'],
            'question_type': self.kwargs['question_type'],
            'pk': self.object.pk
        }
        return reverse('ddm_questionnaire:edit', kwargs=kwargs)

    def get_success_message(self, cleaned_data):
        return self.success_message % dict(
            cleaned_data,
            question_type=self.QUESTION_CLASSES[self.kwargs['question_type']].DEFAULT_QUESTION_TYPE.label
        )


class QuestionEdit(SuccessMessageMixin, DDMAuthMixin, QuestionFormMixin, UpdateView):
    """ View to edit question. """
    model = QuestionBase
    template_name = 'ddm_questionnaire/edit.html'
    success_message = 'Question "%(name)s" was successfully updated.'

    def get_success_url(self):
        success_kwargs = {
            'project_url_id': self.kwargs['project_url_id'],
            'question_type': self.kwargs['question_type'],
            'pk': self.kwargs['pk']
        }
        return reverse('ddm_questionnaire:edit', kwargs=success_kwargs)


class QuestionDelete(SuccessMessageMixin, DDMAuthMixin, ProjectMixin, DeleteView):
    """ View to delete question. """
    model = QuestionBase
    template_name = 'ddm_questionnaire/delete.html'
    success_message = 'Question "%s" was deleted.'

    def get_success_message(self, cleaned_data):
        return self.success_message % self.object.name

    def get_success_url(self):
        return reverse(
        'ddm_questionnaire:overview', kwargs={'project_url_id': self.kwargs['project_url_id']})


class InlineFormsetMixin(ProjectMixin):
    model = QuestionBase
    fields = '__all__'
    formset_model = None
    context_title = None
    fields_to_exclude = ()

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        formset = self.get_formset()
        if formset.is_valid():
            return self.form_valid(formset)
        else:
            return super().form_invalid(formset)

    def get_context_data(self, **kwargs):
        context = super().get_context_data()
        context.update({
            'formset': self.get_formset(),
            'context_title': self.context_title,
            'question': self.object
        })
        return context

    def get_initial_extra_data(self):
        """ Placeholder function to overwrite in views. """
        return []

    def get_formset(self):
        formset = inlineformset_factory(
            QuestionBase, self.formset_model, exclude=self.get_excluded_fields(),
            extra=0
        )
        if self.request.method == "GET":
            initial_data = self.get_initial_extra_data()
        else:
            initial_data = None
        return formset(self.request.POST or None, instance=self.object, initial=initial_data)

    def get_excluded_fields(self):
        if not isinstance(self.object, SemanticDifferential):
            excluded_fields = ('label_alt',)
        else:
            excluded_fields = ()
        return excluded_fields


class ItemEdit(SuccessMessageMixin, DDMAuthMixin, InlineFormsetMixin, UpdateView):
    """ View to edit the items associated with a question. """
    model = QuestionBase
    formset_model = QuestionItem
    template_name = 'ddm_questionnaire/edit_set.html'
    context_title = 'Items'
    success_message = 'Question items updated.'

    def get_success_url(self):
        question = self.get_object()
        success_kwargs = {
            'project_url_id': self.kwargs['project_url_id'],
            'question_type': question.question_type,
            'pk': question.pk
        }
        return reverse('ddm_questionnaire:items', kwargs=success_kwargs)


class ScaleEdit(SuccessMessageMixin, DDMAuthMixin, InlineFormsetMixin, UpdateView):
    """ View to edit the scale associated with a question. """
    model = QuestionBase
    formset_model = ScalePoint
    template_name = 'ddm_questionnaire/edit_set.html'
    context_title = 'Scale Points'
    success_message = 'Question scale updated.'

    def get_success_url(self):
        question = self.get_object()
        success_kwargs = {
            'project_url_id': self.kwargs['project_url_id'],
            'question_type': question.question_type,
            'pk': question.pk
        }
        return reverse('ddm_questionnaire:scale', kwargs=success_kwargs)



from django.urls import path

from ddm.apis import views


app_name = 'ddm_apis'
urlpatterns = [
    path(
        'project/<slug:project_url_id>/overview',
        views.ProjectDetailApiView.as_view(),
        name='project_overview'
    ),
    path(
        'project/<slug:project_url_id>/donations',
        views.ProjectDonationsListView.as_view(),
        name='donations'
    ),
    path(
        'project/<slug:project_url_id>/details/download',
        views.DownloadProjectDetailsView.as_view(),
        name='download_project_details'
    ),
    path(
        'project/<slug:project_url_id>/data/delete',
        views.DeleteProjectData.as_view(),
        name='project_data_delete'
    ),
    path(
        'project/<slug:project_url_id>/responses',
        views.ResponsesApiView.as_view(),
        name='responses'
    ),
    path(
        'project/<slug:project_url_id>/participant/<slug:participant_id>/delete',
        views.DeleteParticipantAPI.as_view(),
        name='participant_delete'
    ),
]



from django.utils import timezone

from ddm.logging.utils import log_server_exception
from ddm.participation.models import Participant
from ddm.projects.models import DonationProject
from ddm.questionnaire.exceptions import QuestionValidationError
from ddm.questionnaire.models import QuestionBase, QuestionnaireResponse


def save_questionnaire_to_db(
        data: dict,
        project: DonationProject,
        participant: Participant) -> None:
    """
    Takes a dictionary object containing information on a participant's
    responses to a project's questions, validates the responses, and saves the
    responses to the database.

    The expected data dictionary is of the form:
    {
      "question_id as str": {
        "response": <response>,
        "question": <question text>,
        "items": [
          {
            "id": <item id as int>,
            "label": <item label>,
            "label_alt": <item alt label>,
            "index": <item index>},
            "value": <value>,
            "randomize": <randomize (boolean)>
          },
          { <next item> }, ...
        ]
      },
      "next question_id as str": { ... },
    }
    """
    for question_id in data:
        try:
            question = QuestionBase.objects.get(pk=int(question_id))
        except QuestionBase.DoesNotExist:
            msg = ('Questionnaire Post Exception:'
                   f'Question with id={question_id} does not exist.')
            log_server_exception(project, msg)
            continue
        except (ValueError, TypeError):
            msg = ('Questionnaire Post Exception: Received invalid '
                   f'question_id ({question_id}) in questionnaire '
                   f'post_data.')
            log_server_exception(project, msg)
            continue

        try:
            question.validate_response(data[question_id]['response'])
        except QuestionValidationError as e:
            # Log validation errors but keep invalid value in data and save to the database.
            for error in e.errors:
                log_server_exception(project, error)

    QuestionnaireResponse.objects.create(
        project=project,
        participant=participant,
        time_submitted=timezone.now(),
        data=data
    )
    return



import csv
import io

from django.core.exceptions import BadRequest, ObjectDoesNotExist
from django.http import Http404, StreamingHttpResponse, HttpResponse
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils import timezone
from django.views.decorators.debug import sensitive_variables
from rest_framework import exceptions, permissions, authentication, status
from rest_framework.generics import ListAPIView
from rest_framework.response import Response
from rest_framework.views import APIView

from ddm.apis.serializers import (
    DataDonationSerializer, ResponseSerializer, ParticipantSerializer,
    ResponseSerializerWithSnapshot, ProjectSerializer
)
from ddm.auth.models import ProjectTokenAuthenticator
from ddm.auth.utils import user_has_project_access
from ddm.datadonation.models import DataDonation
from ddm.encryption.models import Decryption
from ddm.logging.models import EventLogEntry
from ddm.participation.models import Participant
from ddm.projects.models import DonationProject
from ddm.questionnaire.models import QuestionnaireResponse


class DDMAPIMixin:
    """
    Mixin containing ddm-specific methods to be combined with DRF-views.
    """
    def permission_denied(self, request, message=None, code=None):
        """
        If request is not permitted, determine what kind of exception to raise.
        Added EventLog entries.
        """
        if request.authenticators and not request.successful_authenticator:
            self.create_event_log(
                descr='Failed Attempt', msg='Authentication failed.')
            raise exceptions.NotAuthenticated()

        self.create_event_log(
            descr='Failed Attempt', msg='Permission Denied.')
        raise exceptions.PermissionDenied(detail=message, code=code)

    def get_project(self):
        """ Returns project instance. """
        project_id = self.kwargs.get('project_url_id')
        return get_object_or_404(DonationProject, url_id=project_id)

    def create_event_log(self, descr, msg):
        """ Creates an event log entry related to the current project. """
        if self.request is not None:
            prefix = f'{self.request.get_full_path()} {self.request.method}: '
        else:
            prefix = ''
        return EventLogEntry.objects.create(
            project=self.get_project(), description=f'{prefix}{descr}', message=msg)


class ProjectDetailApiView(ListAPIView, DDMAPIMixin):
    """
    API View to retrieve detail information about a specific project.

    This view handles requests to fetch project-related metadata,
    including information participants, blueprint configuration, and project
    configuration.

    Returns:
    - A Response object with data and status code.

    Error Responses:
    - 400 Bad Request: If there's an issue with the input data.
    - 401 Unauthorized: If authentication fails.
    """
    project = None
    authentication_classes = [ProjectTokenAuthenticator]
    permission_classes = [permissions.IsAuthenticated]

    def dispatch(self, request, *args, **kwargs):
        try:
            self.project = self.get_project()
        except Http404 as e:
            raise Http404(str(e))
        return super().dispatch(request, *args, **kwargs)

    def get_participants(self):
        return self.project.participant_set.all()

    def get_blueprints(self):
        return self.project.donationblueprint_set.all()

    def get_blueprints_info(self, blueprints):
        blueprints_info = []
        for blueprint in blueprints:
            blueprints_info.append(blueprint.get_config())
        return blueprints_info

    def get(self, request, format=None, *args, **kwargs):
        participants = self.get_participants()

        blueprints = self.get_blueprints()
        blueprints_info = self.get_blueprints_info(blueprints)

        response = {
            'participants': ParticipantSerializer(participants, many=True).data,
            'blueprints': blueprints_info,
            'project': ProjectSerializer(self.project).data,
            'metadata': {
                'n_blueprints': len(blueprints),
                'n_participants': len(participants)
            }
        }
        self.create_event_log(
            descr='API access: Project details retrieved.',
            msg='The project detail data were accessed through the API.'
        )
        return Response(response)


class ProjectDonationsListView(ListAPIView, DDMAPIMixin):
    """
    API View to retrieve the donations collected in a project.

    Returns:
    - A Response object with data and status code.

    Available query filters:
    - participants (required): A comma separated list of external participant IDs
        belonging to the project whose donations should be returned
        (e.g. ?participants='ID1,ID2,ID3').
        If not specified, no data is returned.
    - blueprints (optional): A comma separated list of file blueprint IDs, for which the
        donations should be returned (e.g. ?blueprints='BP-ID1,BP-ID2,BP-ID3').
        If not specified, donations for all blueprints associated with the
        project are returned.

    Error Responses:
    - 400 Bad Request: If there's an issue with the input data.
    - 401 Unauthorized: If authentication fails.
    - 405 Method Not Allowed: If resources belonging to a super secret project
        are requested.

    Note:
    Please note that this endpoint cannot be used for secret projects.
    """
    authentication_classes = [ProjectTokenAuthenticator]
    permission_classes = [permissions.IsAuthenticated]

    def dispatch(self, request, *args, **kwargs):
        try:
            self.project = self.get_project()
        except Http404 as e:
            raise Http404(str(e))
        return super().dispatch(request, *args, **kwargs)

    def get_participants(self, project):
        participants = project.participant_set.all()
        participant_param = self.request.query_params.get('participants')
        if participant_param is None:
            raise BadRequest('Required parameter "participants" is missing.')
        participant_ids = participant_param.split(',')
        return participants.filter(external_id__in=participant_ids)

    def get_blueprints(self, project):
        blueprints = project.donationblueprint_set.all()
        blueprint_param = self.request.query_params.get('blueprints')
        if blueprint_param is None:
            return blueprints
        else:
            blueprint_ids = blueprint_param.split(',')
            return blueprints.filter(pk__in=blueprint_ids)

    def get_donations(self, blueprint, participants):
        donations = blueprint.datadonation_set.filter(
            participant__in=participants)
        return donations

    def get_metadata(self, project):
        return

    @sensitive_variables()
    def get(self, request, format=None, *args, **kwargs):
        if self.project.super_secret:
            raise exceptions.MethodNotAllowed(
                'GET', detail='Endpoint disabled for super secret projects.')

        participants = self.get_participants(self.project)
        blueprints = self.get_blueprints(self.project)

        response = {
            'blueprints': {},
            'metadata': {
                'n_blueprints': len(blueprints)
            }
        }

        decryptor = Decryption(self.project.secret_key, self.project.get_salt())
        for blueprint in blueprints:
            donations = self.get_donations(blueprint, participants)
            response['blueprints'][f'{blueprint.pk}'] = {
                'blueprint_name': blueprint.name,
                'donations': DataDonationSerializer(donations, many=True, decryptor=decryptor).data
            }

        self.create_event_log(
            descr='API access: Donations retrieved.',
            msg='The donation data was accessed through the API.'
        )
        return Response(response)


class ResponsesApiView(ListAPIView, DDMAPIMixin):
    """
    API View to retrieve the questionnaire responses for a project.

    Returns:
    - A Response object with the data as JSON or CSV and status code.

    Available query filters:
    - participants (optional): A comma-separated list of external participant IDs
        belonging to the project whose donations should be returned
        (e.g. "?participants=ID1,ID2,ID3").
    - include_snapshot (optional): If include_snapshot is 'true', the response
        includes a snapshot of the questionnaire configuration at
        the time the questionnaire was completed for each participant.
        Otherwise, it is omitted.
    - csv (optional): If csv is 'true', the response will be provided
        as a csv file.

    Error Responses:
    - 400 Bad Request: If there's an issue with the input data.
    - 403 Unauthorized: If authentication fails.
    """
    project = None
    authentication_classes = [
        authentication.SessionAuthentication,
        ProjectTokenAuthenticator
    ]
    permission_classes = [permissions.IsAuthenticated]

    def dispatch(self, request, *args, **kwargs):
        try:
            self.project = self.get_project()
        except Http404 as e:
            raise Http404(str(e))
        return super().dispatch(request, *args, **kwargs)

    def get_participants(self):
        participants = self.project.participant_set.all()
        participant_param = self.request.query_params.get('participants')
        if participant_param is None:
            return participants
        else:
            participant_ids = participant_param.split(',')
            return participants.filter(external_id__in=participant_ids)

    def get_responses(self, participants):
        return QuestionnaireResponse.objects.filter(
                project=self.project, participant__in=participants)

    def get_serializer(self):
        snapshot = self.request.query_params.get('include_snapshot')
        if snapshot == 'true':
            return ResponseSerializerWithSnapshot
        else:
            return ResponseSerializer

    def get_metadata(self, responses):
        metadata = {
            'n_responses': len(responses)
        }
        return metadata

    def get_filename(self):
        date = timezone.now().strftime("%Y-%m-%d-%H-%M")
        return f'ddm_{self.project.url_id}_responses_{date}.csv'

    def create_json_response(self, responses, decryptor):
        serializer = self.get_serializer()
        response = {
            'responses': serializer(responses, many=True, decryptor=decryptor).data,
            'metadata': self.get_metadata(responses)
        }
        self.create_event_log(
            descr='API access: Responses retrieved',
            msg='The response data was accessed through the API.'
        )
        return Response(response)

    def create_csv_response(self, responses, decryptor):
        clean_responses = []
        col_names = {'participant', 'time_submitted'}
        responses = ResponseSerializer(responses, many=True, decryptor=decryptor).data
        for response in responses:
            data = dict()
            data['participant'] = response['participant']
            data['time_submitted'] = response['time_submitted']
            for var, val in response['response_data'].items():
                col_names.add(var)
                data[var] = val
            clean_responses.append(data)

        csv_output = io.StringIO()
        writer = csv.DictWriter(csv_output, fieldnames=col_names)
        writer.writeheader()
        writer.writerows(clean_responses)

        csv_content = csv_output.getvalue()
        csv_output.close()

        filename = self.get_filename()
        response = HttpResponse(csv_content, content_type='text/csv')
        response['Content-Disposition'] = f'attachment; filename="{filename}"'
        return response

    def get(self, request, format=None, *args, **kwargs):
        participants = self.get_participants()
        responses = self.get_responses(participants)
        decryptor = Decryption(self.project.secret_key, self.project.get_salt())

        if self.request.query_params.get('csv') == 'true':
            return self.create_csv_response(responses, decryptor)
        else:
            return self.create_json_response(responses, decryptor)


class DeleteParticipantAPI(APIView, DDMAPIMixin):
    """
    Delete a participant by providing their external ID.

    Returns:
    - A Response object with the status code.

    Authentication Methods:
    - Token authentication for remote calls.
    - Session authentication for access through web application (by verifying
        that the requesting user is the project owner).

    Error Responses:
    - 401 Unauthorized: If authentication fails.
    - 403 Forbidden: If a user is not permitted to access a project.
    - 404 Not Found: No participant could be found for the provided ID.
    """
    authentication_classes = [ProjectTokenAuthenticator,
                              authentication.SessionAuthentication]
    permission_classes = [permissions.IsAuthenticated]

    def delete(self, request, format=None, *args, **kwargs):
        """
        Delete participant related to current project by providing external_id.
        """
        project = self.get_project()
        if not user_has_project_access(request.user, project):
            self.create_event_log(
                descr='Forbidden Deletion Request',
                msg='Request user is not permitted to delete the project data.'
            )
            msg = 'User does not have access.'
            return Response(status=status.HTTP_403_FORBIDDEN,
                            data={'message': msg})

        external_id = self.kwargs['participant_id']
        try:
            participant = Participant.objects.get(project=project, external_id=external_id)
        except ObjectDoesNotExist:
            msg = f'Participant with id "{external_id}" not found.'
            return Response(status=status.HTTP_404_NOT_FOUND, data={'message': msg})

        participant.delete()

        self.create_event_log(
            descr=f'Participant with external id "{external_id}" was deleted.',
            msg='Participant deleted.'
        )

        msg = f'Participant with external id "{external_id}" successfully deleted.'
        return Response(status=status.HTTP_200_OK, data={'message': msg})


class DeleteProjectData(APIView, DDMAPIMixin):
    """
    Delete all data (i.e., data donations and questionnaire responses)
    collected for a given donation project.

    Returns:
    - A Response object with the status code.

    Example Usage:
    ```
    DELETE /api/project/<project_pk>/data/delete
    {
        'status': '200',
        'data': {'message': 'Deleted <n donations> data donations and
                            <n responses> questionnaire responses.'}
    }
    ```

    Authentication Methods:
    - Session authentication for access through web application (by verifying
        that the requesting user is the project owner).

    Error Responses:
    - 400 Bad Request: If there's an issue with the input data.
    - 401 Unauthorized: If authentication fails.
    - 403 Forbidden: If a user is not permitted to access a project.

    Note: Description currently not included in the researcher documentation
    because this view can only be called from within the application.
    """
    authentication_classes = [authentication.SessionAuthentication]
    permission_classes = [permissions.IsAuthenticated]

    def delete(self, request, format=None, *args, **kwargs):
        """
        Delete data donations and questionnaire requests associated with project.
        """
        project = self.get_project()
        if not user_has_project_access(request.user, project):
            self.create_event_log(
                descr='Forbidden Deletion Request',
                msg='Request user is not permitted to delete the project data.'
            )
            msg = 'User does not have access.'
            return Response(status=status.HTTP_403_FORBIDDEN,
                            data={'message': msg})

        # Delete all related objects.
        Participant.objects.filter(project=project).delete()
        n_deleted_donations = DataDonation.objects.filter(project=project).delete()[0]
        n_deleted_responses = QuestionnaireResponse.objects.filter(project=project).delete()[0]

        self.create_event_log(
            descr='Data Deletion Successful',
            msg='The project data was deleted.'
        )

        msg = (f'Deleted {n_deleted_donations} data donations and '
               f'{n_deleted_responses} questionnaire responses.')
        return Response(status=status.HTTP_200_OK, data={'message': msg})


class DownloadProjectDetailsView(APIView, DDMAPIMixin):
    """
    Returns a CSV file containing high level participation information for a
    project.

    Each line of the CSV file represents one participant and holds the following
    information:
    - participant ID
    - start time of participation
    - end time of participation
    - current step of participation
    - completed (boolean)
    - extra data collected on participant-level
    - download link for donations
    - for each blueprint:
    -- name
    -- time submitted
    -- consent
    -- status

    Returns a StreamingHttpResponse.
    """
    authentication_classes = [authentication.SessionAuthentication]
    permission_classes = [permissions.IsAuthenticated]

    def dispatch(self, request, *args, **kwargs):
        try:
            self.project = self.get_project()
        except Http404 as e:
            raise Http404(str(e))
        return super().dispatch(request, *args, **kwargs)

    def get_participants(self):
        return self.project.participant_set.all()

    def get_blueprints(self):
        return self.project.donationblueprint_set.all()

    def get_donations(self, participant):
        return participant.datadonation_set.all()

    def get_csv_header(self, blueprints):
        participant_header = self.get_participant_header()
        blueprint_header = self.get_blueprint_header(blueprints)
        return participant_header + blueprint_header + 'donation_download_link\n'

    def get_filename(self):
        date = timezone.now().strftime("%Y-%m-%d-%H-%M")
        return f'ddm_{self.project.url_id}_participation-overview_{date}.csv'

    @staticmethod
    def get_participant_header():
        participant_header = (
            'participant_id,'
            'start_time,'
            'end_time,'
            'current_step,'
            'completed,'
            'extra_data,'
        )
        return participant_header

    @staticmethod
    def get_participant_data(participant):
        participant_info = (
            f'"{participant.external_id}",'
            f'"{participant.start_time}",'
            f'"{participant.end_time}",'
            f'"{participant.current_step}",'
            f'"{participant.completed}",'
            f'"{participant.extra_data}",'
        )
        return participant_info

    @staticmethod
    def get_blueprint_header(blueprints):
        header = ''
        for blueprint in blueprints:
            pk = blueprint.pk
            blueprint_header = (
                f'blueprint_{pk}_name,'
                f'blueprint_{pk}_time_submitted,'
                f'blueprint_{pk}_consent,'
                f'blueprint_{pk}_status,'
            )
            header += blueprint_header
        return header

    def get_blueprint_donation_data(self, blueprints, participant):
        blueprint_data = ''
        n_donations = 0
        donations = self.get_donations(participant)
        for blueprint in blueprints:
            donation = donations.filter(blueprint=blueprint).first()
            if donation:
                data = (
                    f'"{blueprint.name}",'
                    f'"{donation.time_submitted}",'
                    f'"{donation.consent}",'
                    f'"{donation.status}",'
                )
                n_donations += 1
            else:
                data = (
                    f'{None},'
                    f'{None},'
                    f'{None},'
                    f'{None},'
                )
            blueprint_data += data

        if n_donations > 0:
            blueprint_data += self.get_download_link(participant)
        else:
            blueprint_data += f'{None}'
        return blueprint_data

    def get_download_link(self, participant):
        url = reverse(
            'ddm_datadonation:download_donation',
            args=[self.project.url_id, participant.external_id]
        )
        return self.request.build_absolute_uri(url)

    def generate_csv(self):
        blueprints = self.get_blueprints()
        header = self.get_csv_header(blueprints)
        yield header

        # Loop through participants
        participants = self.get_participants()
        n_rows = 0
        data = ''
        for participant in participants:
            participant_data = self.get_participant_data(participant)
            blueprint_data = self.get_blueprint_donation_data(blueprints, participant)
            data += (participant_data + blueprint_data + '\n')
            n_rows += 1
            if n_rows % 10 == 0:
                yield data
                data = ''
        yield data

    def get(self, request, *args, **kwargs):
        """
        Handle GET request to stream JSON response.
        """
        if (not user_has_project_access(request.user, self.project) and
                not request.user.is_superuser):
            self.create_event_log(
                descr='Forbidden Data Overview Request',
                msg='Request user is not permitted to get the project data.'
            )
            msg = 'User does not have access.'
            raise exceptions.PermissionDenied(msg)

        response = StreamingHttpResponse(
            self.generate_csv(),
            content_type='text/csv'
        )
        filename = self.get_filename()
        response['Content-Disposition'] = f'attachment; filename="{filename}"'
        return response



import io
import json
import zipfile

from django.contrib.auth import get_user_model
from django.contrib.sessions.middleware import SessionMiddleware
from django.test import Client, TestCase, override_settings, RequestFactory
from django.urls import reverse
from django.utils import timezone

from ddm.datadonation.models import DataDonation, DonationBlueprint, FileUploader
from ddm.logging.models import ExceptionLogEntry
from ddm.participation.models import Participant
from ddm.participation.views import (
    DataDonationView, ContinuationView, DebriefingView,
    QuestionnaireView, BriefingView
)
from ddm.projects.models import DonationProject, ResearchProfile
from ddm.questionnaire.models import OpenQuestion


User = get_user_model()


@override_settings(DDM_SETTINGS={'EMAIL_PERMISSION_CHECK':  r'.*(\.|@)mail\.com$', })
class ParticipationFlowBaseTestCase(TestCase):

    @classmethod
    def setUpTestData(cls):
        user = User.objects.create_user(**{
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        })
        profile = ResearchProfile.objects.create(user=user)

        cls.project_base = DonationProject.objects.create(
            name='Base Project', slug='base', owner=profile)

        cls.project_alt = DonationProject.objects.create(
            name='Alt Project', slug='alt', owner=profile)

        file_uploader = FileUploader.objects.create(
            project=cls.project_base,
            name='basic file uploader',
            upload_type=FileUploader.UploadTypes.SINGLE_FILE
        )

        cls.blueprint = DonationBlueprint.objects.create(
            project=cls.project_base,
            name='donation blueprint',
            expected_fields='"a", "b"',
            file_uploader=file_uploader
        )

        OpenQuestion.objects.create(
            project=cls.project_base,
            blueprint=cls.blueprint,
            name='open question',
            variable_name='open_question'
        )

        # URLs for project with questionnaire.
        slug_base = cls.project_base.slug
        cls.briefing_url = reverse(
            'ddm_participation:briefing', args=[slug_base])
        cls.dd_url = reverse(
            'ddm_participation:datadonation', args=[slug_base])
        cls.quest_url = reverse(
            'ddm_participation:questionnaire', args=[slug_base])
        cls.debriefing_url = reverse(
            'ddm_participation:debriefing', args=[slug_base])

        cls.inactive_info_page = reverse(
            'ddm_participation:project_inactive', args=[slug_base])

        # URLs for project without questionnaire.
        slug_alt = cls.project_alt.slug
        cls.briefing_url_no_quest = reverse(
            'ddm_participation:briefing', args=[slug_alt])
        cls.dd_url_no_quest = reverse(
            'ddm_participation:datadonation', args=[slug_alt])
        cls.quest_url_no_quest = reverse(
            'ddm_participation:questionnaire', args=[slug_alt])
        cls.debriefing_url_no_quest = reverse(
            'ddm_participation:debriefing', args=[slug_alt])

        # URLs for non-existing project.
        cls.briefing_url_invalid = reverse(
            'ddm_participation:briefing', args=['nope'])
        cls.dd_url_invalid = reverse(
            'ddm_participation:datadonation', args=['nope'])
        cls.quest_url_invalid = reverse(
            'ddm_participation:questionnaire', args=['nope'])
        cls.debriefing_url_invalid = reverse(
            'ddm_participation:debriefing', args=['nope'])

    def initialize_project_and_session(self):
        self.client = Client()
        self.client.get(self.briefing_url)
        self.client.get(self.briefing_url_no_quest)

    def get_participant(self, project_id):
        session = self.client.session
        participant_id = session[f'project-{project_id}']['participant_id']
        participant = Participant.objects.get(pk=int(participant_id))
        return participant

    def create_data_donation(self):
        session = self.client.session
        participant_id = session[f'project-{self.project_base.pk}']['participant_id']
        participant = Participant.objects.get(pk=int(participant_id))
        DataDonation.objects.create(
            project=self.project_base,
            blueprint=self.blueprint,
            participant=participant,
            time_submitted=timezone.now(),
            consent=True,
            status='success',
            data='{}'
        )


class TestBriefingView(ParticipationFlowBaseTestCase):

    def setUp(self):
        super().setUp()

    def test_project_briefing_view_registers_project(self):
        self.client.get(self.briefing_url)
        session = self.client.session
        expected_keys = ['participant_id']
        actual_keys = set(session[f'project-{self.project_base.pk}'].keys())
        assert set(expected_keys).issubset(actual_keys)

    def test_project_briefing_view_registers_participant(self):
        nr_participants_before = len(Participant.objects.all())
        self.client.get(self.briefing_url)
        self.assertGreater(len(Participant.objects.all()), nr_participants_before)

        project_session = self.client.session[f'project-{self.project_base.pk}']
        self.assertIsNotNone(project_session['participant_id'])

    def test_project_briefing_view_extracts_parameter(self):
        self.project_base.expected_url_parameters = 'testparam'
        self.project_base.url_parameter_enabled = True
        self.project_base.save()
        self.client.get(self.briefing_url + '?testparam=okay&altparam=false')
        project_session = self.client.session[f'project-{self.project_base.pk}']
        participant_id = project_session['participant_id']
        participant = Participant.objects.get(pk=participant_id)
        self.assertIn('testparam', participant.extra_data['url_param'])
        self.assertNotIn('altparam', participant.extra_data['url_param'])
        self.assertEqual(participant.extra_data['url_param']['testparam'], 'okay')

    def test_project_briefing_view_GET_valid_url(self):
        response = self.client.get(self.briefing_url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, BriefingView.template_name)

    def test_project_briefing_view_GET_invalid_url(self):
        response = self.client.get(self.briefing_url_invalid, follow=True)
        self.assertEqual(response.status_code, 404)

    def test_project_briefing_POST(self):
        response = self.client.post(self.briefing_url)
        participant = self.get_participant(self.project_base.pk)

        self.assertEqual(participant.current_step, 1)
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, self.dd_url)

    def test_project_briefing_consent_yes_POST(self):
        self.project_base.briefing_consent_enabled = True
        self.project_base.save()
        self.client.get(self.briefing_url)
        response = self.client.post(
            self.briefing_url, {'briefing_consent': '1'}, follow=True)

        participant = self.get_participant(self.project_base.pk)
        self.assertEqual(participant.current_step, 1)
        self.assertRedirects(response, self.dd_url)
        self.assertEqual(participant.extra_data['briefing_consent'], '1')

    def test_project_briefing_consent_no_POST(self):
        self.project_base.briefing_consent_enabled = True
        self.project_base.save()
        self.client.get(self.briefing_url)
        response = self.client.post(self.briefing_url, {'briefing_consent': '0'})

        participant = self.get_participant(self.project_base.pk)
        self.assertEqual(participant.current_step, 3)
        self.assertRedirects(response, self.debriefing_url)
        self.assertEqual(participant.extra_data['briefing_consent'], '0')

    def test_project_briefing_consent_invalid_POST(self):
        self.project_base.briefing_consent_enabled = True
        self.project_base.save()
        self.client.get(self.briefing_url)
        response = self.client.post(self.briefing_url)

        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, BriefingView.template_name)

    def test_project_inactive_redirects_to_info_page(self):
        self.project_base.active = False
        self.project_base.save()

        response = self.client.get(self.briefing_url, follow=True)
        self.assertRedirects(response, self.inactive_info_page)

        self.project_base.active = True
        self.project_base.save()

class TestDonationView(ParticipationFlowBaseTestCase):

    def setUp(self):
        super().setUp()
        self.initialize_project_and_session()
        self.create_data_donation()
        participant = self.get_participant(self.project_base.pk)
        participant.current_step = 1
        participant.save()

    def initialize_view(self):
        request = RequestFactory().get('/sadf/<slug:slug>/')
        middleware = SessionMiddleware(lambda req: None)
        middleware.process_request(request)
        request.session.save()
        view = DataDonationView()
        view.setup(request, **{'slug': self.project_base.slug})
        return view

    @staticmethod
    def get_zip_file(file_name, file_content):
        zip_buffer = io.BytesIO()
        with zipfile.ZipFile(zip_buffer, 'w', zipfile.ZIP_DEFLATED) as zip_file:
            try:
                zip_file.writestr(file_name, file_content.encode('utf-8'))
            except UnicodeEncodeError:
                zip_file.writestr(file_name, file_content.encode('utf-8', errors='replace'))
        zip_buffer.seek(0)
        return zip_buffer

    def test_data_donation_GET_valid_url(self):
        response = self.client.get(self.dd_url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, DataDonationView.template_name)

    def test_data_donation_GET_invalid_url(self):
        response = self.client.get(self.dd_url_invalid, follow=True)
        self.assertEqual(response.status_code, 404)

    def test_data_donation_POST_redirect(self):
        response = self.client.post(self.dd_url)
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, self.quest_url)

    def test_process_uploads_invalid_post_data(self):
        view = self.initialize_view()
        files = {}
        exceptions_count_before = ExceptionLogEntry.objects.count()
        view.process_uploads(files)
        exceptions_count_after = ExceptionLogEntry.objects.count()
        last_exception = ExceptionLogEntry.objects.order_by('date').last()
        self.assertEqual(exceptions_count_before, (exceptions_count_after - 1))
        self.assertIn('not receive expected data file from client',
                      last_exception.message)

    def test_process_uploads_non_zip_file(self):
        view = self.initialize_view()
        files = {'post_data': 'This is not a zip file.'}
        exceptions_count_before = ExceptionLogEntry.objects.count()
        view.process_uploads(files)
        exceptions_count_after = ExceptionLogEntry.objects.count()
        last_exception = ExceptionLogEntry.objects.order_by('date').last()
        self.assertEqual(exceptions_count_before, (exceptions_count_after - 1))
        self.assertIn('not a zip file', last_exception.message)

    def test_process_uploads_zip_file_with_missing_content(self):
        view = self.initialize_view()
        zip_buffer = self.get_zip_file('invalid_name.json', '{}')
        files = {'post_data': zip_buffer}
        exceptions_count_before = ExceptionLogEntry.objects.count()
        view.process_uploads(files)
        exceptions_count_after = ExceptionLogEntry.objects.count()
        last_exception = ExceptionLogEntry.objects.order_by('date').last()
        self.assertEqual(exceptions_count_before, (exceptions_count_after - 1))
        self.assertIn('"ul_data.json" is not in namelist',
                      last_exception.message)

    def test_process_uploads_invalid_encoding(self):
        view = self.initialize_view()
        invalid_string = 'Hello, 你好, مرحبا, \ud83d'
        zip_buffer = self.get_zip_file('ul_data.json', invalid_string)
        files = {'post_data': zip_buffer}
        exceptions_count_before = ExceptionLogEntry.objects.count()
        view.process_uploads(files)
        exceptions_count_after = ExceptionLogEntry.objects.count()
        last_exception = ExceptionLogEntry.objects.order_by('date').last()
        self.assertEqual(exceptions_count_before, (exceptions_count_after - 1))
        self.assertIn('decode error in donated data',
                      last_exception.message)

    def test_process_uploads_json_error(self):
        view = self.initialize_view()
        zip_buffer = self.get_zip_file('ul_data.json', '{12: "some data"}')
        files = {'post_data': zip_buffer}
        exceptions_count_before = ExceptionLogEntry.objects.count()
        view.process_uploads(files)
        exceptions_count_after = ExceptionLogEntry.objects.count()
        last_exception = ExceptionLogEntry.objects.order_by('date').last()
        self.assertEqual(exceptions_count_before, (exceptions_count_after - 1))
        self.assertIn('decode error in donated data',
                      last_exception.message)

    def test_process_uploads_blueprint_does_not_exist(self):
        view = self.initialize_view()
        zip_buffer = self.get_zip_file('ul_data.json', '{"12": "some data"}')
        files = {'post_data': zip_buffer}
        exceptions_count_before = ExceptionLogEntry.objects.count()
        view.process_uploads(files)
        exceptions_count_after = ExceptionLogEntry.objects.count()
        last_exception = ExceptionLogEntry.objects.order_by('date').last()
        self.assertEqual(exceptions_count_before, (exceptions_count_after - 1))
        self.assertIn('blueprint with id=', last_exception.message)

    def test_process_uploads_valid_post_data(self):
        view = self.initialize_view()
        valid_data = {
            f'{self.blueprint.pk}' : {
                'consent': True,
                'extracted_data': [],
                'status': 'complete'
            }
        }
        zip_buffer = self.get_zip_file('ul_data.json', json.dumps(valid_data))
        files = {'post_data': zip_buffer}
        exceptions_count_before = ExceptionLogEntry.objects.count()
        donation_count_before = DataDonation.objects.count()
        view.process_uploads(files)
        exceptions_count_after = ExceptionLogEntry.objects.count()
        donation_count_after = DataDonation.objects.count()
        self.assertEqual(exceptions_count_before, exceptions_count_after)
        self.assertEqual(donation_count_before + 1, donation_count_after)

    def test_project_inactive_redirects_to_info_page(self):
        self.project_base.active = False
        self.project_base.save()

        response = self.client.get(self.dd_url)
        self.assertRedirects(response, self.inactive_info_page)

        self.project_base.active = True
        self.project_base.save()

class TestQuestionnaireView(ParticipationFlowBaseTestCase):

    def setUp(self):
        super().setUp()
        self.initialize_project_and_session()
        self.create_data_donation()

        participant_base = self.get_participant(self.project_base.pk)
        participant_base.current_step = 2
        participant_base.save()

        participant_alt = self.get_participant(self.project_alt.pk)
        participant_alt.current_step = 2
        participant_alt.save()

    def test_questionnaire_GET_valid_url(self):
        response = self.client.get(self.quest_url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, QuestionnaireView.template_name)

    def test_questionnaire_GET_invalid_url(self):
        response = self.client.get(self.quest_url_invalid, follow=True)
        self.assertEqual(response.status_code, 404)

    def test_questionnaire_GET_no_questionnaire(self):
        response = self.client.get(self.quest_url_no_quest)
        self.assertRedirects(response, self.debriefing_url_no_quest)

    def test_questionnaire_POST_redirect(self):
        response = self.client.post(self.quest_url)
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, self.debriefing_url)

    def test_project_inactive_redirects_to_info_page(self):
        self.project_base.active = False
        self.project_base.save()

        response = self.client.get(self.quest_url)
        self.assertRedirects(response, self.inactive_info_page)

        self.project_base.active = True
        self.project_base.save()

class TestDebriefingView(ParticipationFlowBaseTestCase):

    def setUp(self):
        super().setUp()
        self.initialize_project_and_session()

        participant = self.get_participant(self.project_base.pk)
        participant.current_step = 3
        participant.save()

    def test_project_debriefing_GET_valid_url(self):
        response = self.client.get(self.debriefing_url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, DebriefingView.template_name)

    def test_project_debriefing_GET_invalid_url(self):
        response = self.client.get(self.debriefing_url_invalid, follow=True)
        self.assertEqual(response.status_code, 404)

    def test_project_inactive_redirects_to_info_page(self):
        self.project_base.active = False
        self.project_base.save()

        response = self.client.get(self.debriefing_url)
        self.assertRedirects(response, self.inactive_info_page)

        self.project_base.active = True
        self.project_base.save()

class TestRedirect(ParticipationFlowBaseTestCase):

    def setUp(self):
        super().setUp()
        self.initialize_project_and_session()
        self.project_base.redirect_enabled = True
        self.project_base.url_parameter_enabled = True
        self.project_base.save()

    def test_redirect_single_parameter(self):
        self.project_base.expected_url_parameters = 'testparam'
        self.project_base.redirect_target = 'http://test.test/?para={{url_parameter.testparam}}'
        self.project_base.save()

        self.client.get(self.briefing_url + '?testparam=test')

        participant = self.get_participant(self.project_base.pk)
        participant.current_step = 3
        participant.save()

        response = self.client.get(self.debriefing_url)
        self.assertEqual(
            response.context['redirect_target'],
            'http://test.test/?para=test'
        )

    def test_redirect_multiple_parameters(self):
        self.project_base.expected_url_parameters = 'testparam;testparam2'
        self.project_base.redirect_target = (
            'http://test.test/?para={{url_parameter.testparam}}&'
            'para2={{url_parameter.testparam2}}'
        )
        self.project_base.save()
        self.client.get(self.briefing_url + '?testparam=test&testparam2=test2')

        participant = self.get_participant(self.project_base.pk)
        participant.current_step = 3
        participant.save()

        response = self.client.get(self.debriefing_url)
        self.assertEqual(
            response.context['redirect_target'],
            'http://test.test/?para=test&para2=test2'
        )


class TestContinuationView(ParticipationFlowBaseTestCase):

    def setUp(self):
        super().setUp()
        self.initialize_project_and_session()

    def test_render_without_participant(self):
        new_client = Client()
        url = reverse(
            'ddm_participation:continuation',
            args=[self.project_base.slug]
        )
        response = new_client.get(url, data={'p': 'some-non-existing-id'}, follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, ContinuationView.template_name)

    def test_redirect_continuation(self):
        self.client.get(self.briefing_url)
        self.client.post(self.briefing_url)
        self.client.get(self.dd_url)
        participant = self.get_participant(self.project_base.pk)
        url = reverse(
            'ddm_participation:continuation',
            args=[self.project_base.slug]
        )
        new_client = Client()
        response = new_client.get(url, data={'p': participant.external_id}, follow=True)
        self.assertRedirects(response, self.dd_url)


class TestParticipationRedirectView(ParticipationFlowBaseTestCase):

    def setUp(self):
        super().setUp()
        self.initialize_project_and_session()

    def test_url_without_step_redirects_to_briefing(self):
        url_input = reverse(
            'ddm_participation:redirect',
            args=[self.project_base.slug]
        )
        url_expected = reverse(
            'ddm_participation:briefing',
            args=[self.project_base.slug]
        )
        new_client = Client()
        response = new_client.get(url_input, follow=True)
        self.assertRedirects(response, url_expected)

    def test_url_without_step_redirects_to_briefing_with_query_string(self):
        query_string = '?param1=A&param2=B'
        url_input = reverse(
            'ddm_participation:redirect',
            args=[self.project_base.slug]
        ) + query_string
        url_expected = reverse(
            'ddm_participation:briefing',
            args=[self.project_base.slug]
        ) + query_string
        new_client = Client()
        response = new_client.get(url_input, follow=True)
        self.assertRedirects(response, url_expected)



# Generated by Django 3.2.13 on 2024-10-26 08:48

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('ddm_projects', '0001_initial'),
        ('ddm_datadonation', '0002_alter_datadonation_participant'),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.AlterField(
                    model_name='datadonation',
                    name='project',
                    field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm_projects.donationproject'),
                ),
                migrations.AlterField(
                    model_name='donationblueprint',
                    name='project',
                    field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm_projects.donationproject'),
                ),
                migrations.AlterField(
                    model_name='fileuploader',
                    name='project',
                    field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm_projects.donationproject'),
                ),
            ],
            # Reusing an existing table, so do nothing.
            database_operations=[]
        )
    ]



# Generated by Django 3.2.13 on 2024-10-26 08:48

import ddm.projects.models
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.CreateModel(
                    name='ResearchProfile',
                    fields=[
                        ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                        ('active', models.BooleanField(default=True)),
                        ('created', models.DateTimeField(default=django.utils.timezone.now, verbose_name='date registered')),
                        ('ignore_email_restriction', models.BooleanField(default=False)),
                        ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='ddm_research_profile', to=settings.AUTH_USER_MODEL)),
                    ],
                ),
                migrations.CreateModel(
                    name='DonationProject',
                    fields=[
                        ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                        ('name', models.CharField(help_text='Project Name - for internal organisation only (can still be changed later).', max_length=50)),
                        ('date_created', models.DateTimeField(default=django.utils.timezone.now)),
                        ('contact_information', models.TextField(help_text='Please provide the contact information of the person responsible for the conduction of this study (Name, professional address, e-mail, tel, ...). The contact information will be accessible for participants during the data donation. The field is mandatory.', null=True, verbose_name='Contact information')),
                        ('data_protection_statement', models.TextField(help_text='Please provide a data protection statement for your data donation collection. This should include the purpose for which the data is collected, how it will be stored, and who will have access to the data. The field is mandatory.', null=True, verbose_name='Data Protection Statement')),
                        ('slug', models.SlugField(help_text='Identifier that is included in the URL through which participants can access the project (e.g, https://root.url/url-identifier).', unique=True, verbose_name='URL Identifier')),
                        ('briefing_text', models.TextField(blank=True, help_text='This text will be displayed to participants in the first step before beginning the data donation. Here, you should introduce your study, who is responsible for the study, provide a quick summary of what your participants will be asked to do in the next steps, etc.', null=True, verbose_name='Briefing Text')),
                        ('briefing_consent_enabled', models.BooleanField(default=False, help_text='Enable this option to obtain explicit consent from participant, that they want to start the study.', verbose_name='Briefing Consent Mandatory')),
                        ('briefing_consent_label_yes', models.CharField(blank=True, max_length=255)),
                        ('briefing_consent_label_no', models.CharField(blank=True, max_length=255)),
                        ('debriefing_text', models.TextField(blank=True, help_text='This text will be displayed to participants on the last page of the study (i.e., after the data donation and the questionnaire).', null=True, verbose_name='Debriefing Text')),
                        ('img_header_left', models.ImageField(blank=True, null=True, upload_to=ddm.projects.models.project_header_dir_path, verbose_name='Header Image Left')),
                        ('img_header_right', models.ImageField(blank=True, null=True, upload_to=ddm.projects.models.project_header_dir_path, verbose_name='Header Image Right')),
                        ('public_key', models.BinaryField()),
                        ('super_secret', models.BooleanField(default=False)),
                        ('redirect_enabled', models.BooleanField(default=False, verbose_name='Redirect enabled')),
                        ('redirect_target', models.CharField(blank=True, help_text='Always include <i>http://</i> or <i>https://</i> in the redirect target. If URL parameter extraction is enabled for this project, you can include the extracted URL parameters in the redirect target as follows: "https://redirect.me/?redirectpara=<b>{{participant.data.url_param.URLParameter}}</b>".', max_length=2000, verbose_name='Redirect target')),
                        ('url_parameter_enabled', models.BooleanField(default=False, verbose_name='URL parameter extraction enabled')),
                        ('expected_url_parameters', models.CharField(blank=True, help_text='Separate multiple parameters with a semikolon (";"). Semikolons are not allowed as part of the expected url parameters.', max_length=500, verbose_name='Expected URL parameter')),
                        ('owner', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='project_owner', to='ddm_projects.researchprofile')),
                    ],
                ),
            ],
            # Tables already exist. See ddm/migrations/0056.
            database_operations=[]
        )
    ]



from django.contrib.auth import get_user_model
from django.test import TestCase

from ddm.projects.models import DonationProject, ResearchProfile
from ddm.questionnaire.exceptions import QuestionValidationError
from ddm.questionnaire.models import (
    SingleChoiceQuestion, QuestionItem, MultiChoiceQuestion, MatrixQuestion,
    ScalePoint, SemanticDifferential
)

User = get_user_model()


class TestQuestionModelsBaseCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        user = User.objects.create_user(**{
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        })
        profile = ResearchProfile.objects.create(user=user)

        cls.project = DonationProject.objects.create(
            name='Base Project', slug='base', owner=profile)

        cls.question_config = {
            'project': cls.project,
            'name': 'Test Question',
            'page': 1,
            'index': 1,
            'variable_name': 'test_var',
            'text': 'Question Text',
        }


class TestSingleChoiceQuestion(TestQuestionModelsBaseCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.question = SingleChoiceQuestion.objects.create(**cls.question_config)
        QuestionItem.objects.create(question=cls.question, index=1, value=1)
        QuestionItem.objects.create(question=cls.question, index=2, value=8)

    def test_validate_response_valid_case(self):
        valid_values = [1, 8, -99, '1', '8', '-99']
        for value in valid_values:
            self.assertTrue(self.question.validate_response(value))

    def test_validate_response_invalid_case(self):
        invalid_values = [0, 1.4, '4', '1.7', 'abc']
        for value in invalid_values:
            with self.assertRaises(QuestionValidationError):
                self.question.validate_response(value)


class TestMultiChoiceQuestion(TestQuestionModelsBaseCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.question = MultiChoiceQuestion.objects.create(**cls.question_config)
        cls.item_a = QuestionItem.objects.create(
            question=cls.question, index=1, value=1)
        cls.item_b = QuestionItem.objects.create(
            question=cls.question, index=2, value=8)

    def test_validate_response_valid_case(self):
        valid_values = [
            {str(self.item_a.pk): 1, str(self.item_b.pk): 0},
            {self.item_a.pk: False, str(self.item_b.pk): '-99'},
            {self.item_a.pk: True, self.item_b.pk: -99},
        ]
        for value in valid_values:
            self.assertTrue(self.question.validate_response(value))

    def test_validate_response_invalid_case(self):
        invalid_values = [
            {str(self.item_a.pk): 1},
            {self.item_a.pk: 'False', str(self.item_b.pk): 0},
            {self.item_a.pk: 1, self.item_b.pk: 1, '3': 1},
        ]
        for value in invalid_values:
            with self.assertRaises(QuestionValidationError):
                self.question.validate_response(value)


class TestMatrixQuestion(TestQuestionModelsBaseCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.question = MatrixQuestion.objects.create(**cls.question_config)
        cls.item_a = QuestionItem.objects.create(
            question=cls.question, index=1, value=1)
        cls.item_b = QuestionItem.objects.create(
            question=cls.question, index=2, value=8)
        cls.scale_a = ScalePoint.objects.create(
            question=cls.question, index=1, value=1)
        cls.scale_b = ScalePoint.objects.create(
            question=cls.question, index=2, value=6)

    def test_validate_response_valid_case(self):
        valid_values = [
            {str(self.item_a.pk): 1, str(self.item_b.pk): 6},
            {self.item_a.pk: '1', str(self.item_b.pk): '-99'},
            {self.item_a.pk: '6', self.item_b.pk: -99},
        ]
        for value in valid_values:
            self.assertTrue(self.question.validate_response(value))

    def test_validate_response_invalid_case(self):
        invalid_values = [
            {str(self.item_a.pk): 1},
            {self.item_a.pk: 1, str(self.item_b.pk): 0},
            {self.item_a.pk: 1, self.item_b.pk: 1, '3': 1},
        ]
        for value in invalid_values:
            with self.assertRaises(QuestionValidationError):
                self.question.validate_response(value)


class TestSemanticDifferentialQuestion(TestQuestionModelsBaseCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.question = SemanticDifferential.objects.create(**cls.question_config)
        cls.item_a = QuestionItem.objects.create(
            question=cls.question, index=1, value=1)
        cls.item_b = QuestionItem.objects.create(
            question=cls.question, index=2, value=8)
        cls.scale_a = ScalePoint.objects.create(
            question=cls.question, index=1, value=1)
        cls.scale_b = ScalePoint.objects.create(
            question=cls.question, index=2, value=6)

    def test_validate_response_valid_case(self):
        valid_values = [
            {str(self.item_a.pk): 1, str(self.item_b.pk): 6},
            {self.item_a.pk: '1', str(self.item_b.pk): '-99'},
            {self.item_a.pk: '6', self.item_b.pk: -99},
        ]
        for value in valid_values:
            self.assertTrue(self.question.validate_response(value))

    def test_validate_response_invalid_case(self):
        invalid_values = [
            {str(self.item_a.pk): 1},
            {self.item_a.pk: 1, str(self.item_b.pk): 0},
            {self.item_a.pk: 1, self.item_b.pk: 1, '3': 1},
        ]
        for value in invalid_values:
            with self.assertRaises(QuestionValidationError):
                self.question.validate_response(value)

    def test_create_config(self):
        expected_item_config = []
        for item in [self.item_a, self.item_b]:
            expected_item_config.append({
                'id': item.pk,
                'label': item.label,
                'label_alt': item.label_alt,
                'index': item.index,
                'value': item.value,
                'randomize': item.randomize,
            })
        expected_scale_config = []
        for scale_point in [self.scale_a, self.scale_b]:
            expected_scale_config.append({
                'id': scale_point.pk,
                'label': scale_point.label,
                'index': scale_point.index,
                'value': scale_point.value,
                'add_border': scale_point.add_border,
            })
        expected_config = {
            'question': self.question.pk,
            'type': self.question.question_type,
            'page': self.question.page,
            'index': self.question.index,
            'text': self.question.text,
            'required': self.question.required,
            'items': expected_item_config,
            'scale': expected_scale_config,
            'options': {}
        }
        self.assertDictEqual(expected_config, self.question.create_config())



from django.urls import path

from ddm.participation import views


app_name = 'ddm_participation'
urlpatterns = [
    path(r'', views.participation_redirect_view, name='redirect'),
    path(r'briefing/', views.BriefingView.as_view(), name='briefing'),
    path(r'data-donation/', views.DataDonationView.as_view(), name='datadonation'),
    path(r'questionnaire/', views.QuestionnaireView.as_view(), name='questionnaire'),
    path(r'debriefing/', views.DebriefingView.as_view(), name='debriefing'),
    path(r'continue/', views.ContinuationView.as_view(), name='continuation'),
    path(r'inactive/', views.ProjectInactiveView.as_view(), name='project_inactive')
]



# Generated by Django 4.2.16 on 2025-02-10 08:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm_projects', '0004_donationproject_url_id'),
    ]

    operations = [
        migrations.AlterField(
            model_name='donationproject',
            name='slug',
            field=models.SlugField(help_text='Identifier that is included in the URL through which participants can access the project (e.g, https://root.url/<b>my-url-identifier</b>). Can only contain letters, hyphens, numbers or underscores.', unique=True, verbose_name='URL Identifier'),
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='url_id',
            field=models.SlugField(max_length=8, unique=True),
        ),
    ]



from django.urls import path

from ddm.projects import views


app_name = 'ddm_projects'
urlpatterns = [
    path(r'', views.ProjectList.as_view(), name='list'),
    path(r'create/', views.ProjectCreate.as_view(), name='create'),
    path(r'<slug:project_url_id>/', views.ProjectDetail.as_view(), name='detail'),
    path(r'<slug:project_url_id>/edit/', views.ProjectEdit.as_view(), name='edit'),
    path(r'<slug:project_url_id>/delete/', views.ProjectDelete.as_view(), name='delete'),
    path(r'<slug:project_url_id>/briefing/', views.BriefingEdit.as_view(), name='briefing_edit'),
    path(r'<slug:project_url_id>/debriefing/', views.DebriefingEdit.as_view(), name='debriefing_edit'),
]



# Generated by Django 3.2.13 on 2024-10-25 10:09

import ddm.questionnaire.models
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('ddm', '0051_auto_20241025_1209'),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.CreateModel(
                    name='QuestionBase',
                    fields=[
                        ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                        ('question_type', models.CharField(choices=[('generic', 'Generic Question'), ('single_choice', 'Single Choice Question'), ('multi_choice', 'Multi Choice Question'), ('matrix', 'Matrix Question'), ('semantic_diff', 'Semantic Differential'), ('open', 'Open Question'), ('transition', 'Text Block')], default='generic', max_length=20)),
                        ('name', models.CharField(max_length=255)),
                        ('page', models.PositiveIntegerField(default=1, verbose_name='Page')),
                        ('index', models.PositiveIntegerField(default=1, verbose_name='Index')),
                        ('variable_name', models.SlugField(help_text='Will be used in the data export to identify responses to this question.', verbose_name='Variable name for storing response')),
                        ('text', models.TextField(blank=True, help_text='If a question is linked to a File Blueprint, data points from the donated data associated with the linked donation blueprint can be included in the question text. This data can be included as "{{ data }}" in the question text. It is possible to subset the data object (e.g., to include the last datapoint you can use {{ data.0 }} or include advanced rendering options included in the Django templating engine. For a more comprehensive overview and examples see the documentation. Additionally, information directly related to the participant can be included in the question text. This information can be referenced as "{{ participant }}".', null=True)),
                        ('required', models.BooleanField(default=False)),
                        ('blueprint', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='ddm.donationblueprint')),
                        ('polymorphic_ctype', models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='polymorphic_ddm_questionnaire.questionbase_set+', to='contenttypes.contenttype')),
                        ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.donationproject')),
                    ],
                    options={
                        'ordering': ['page', 'index'],
                    },
                ),
                migrations.CreateModel(
                    name='MatrixQuestion',
                    fields=[
                        ('questionbase_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='ddm_questionnaire.questionbase')),
                        ('randomize_items', models.BooleanField(default=False)),
                    ],
                    options={
                        'abstract': False,
                    },
                    bases=(ddm.questionnaire.models.ScaleMixin, 'ddm_questionnaire.questionbase', models.Model),
                ),
                migrations.CreateModel(
                    name='MultiChoiceQuestion',
                    fields=[
                        ('questionbase_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='ddm_questionnaire.questionbase')),
                        ('randomize_items', models.BooleanField(default=False)),
                    ],
                    options={
                        'abstract': False,
                    },
                    bases=('ddm_questionnaire.questionbase', models.Model),
                ),
                migrations.CreateModel(
                    name='OpenQuestion',
                    fields=[
                        ('questionbase_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='ddm_questionnaire.questionbase')),
                        ('display', models.CharField(choices=[('small', 'Small'), ('large', 'Large')], default='large', help_text='"Small" displays a one-line textfield, "Large" a multiline textfield as input.', max_length=20)),
                    ],
                    options={
                        'abstract': False,
                        'base_manager_name': 'objects',
                    },
                    bases=('ddm_questionnaire.questionbase',),
                ),
                migrations.CreateModel(
                    name='SemanticDifferential',
                    fields=[
                        ('questionbase_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='ddm_questionnaire.questionbase')),
                        ('randomize_items', models.BooleanField(default=False)),
                    ],
                    options={
                        'abstract': False,
                    },
                    bases=(ddm.questionnaire.models.ScaleMixin, 'ddm_questionnaire.questionbase', models.Model),
                ),
                migrations.CreateModel(
                    name='SingleChoiceQuestion',
                    fields=[
                        ('questionbase_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='ddm_questionnaire.questionbase')),
                        ('randomize_items', models.BooleanField(default=False)),
                    ],
                    options={
                        'abstract': False,
                    },
                    bases=('ddm_questionnaire.questionbase', models.Model),
                ),
                migrations.CreateModel(
                    name='Transition',
                    fields=[
                        ('questionbase_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='ddm_questionnaire.questionbase')),
                    ],
                    options={
                        'abstract': False,
                        'base_manager_name': 'objects',
                    },
                    bases=('ddm_questionnaire.questionbase',),
                ),
                migrations.CreateModel(
                    name='ScalePoint',
                    fields=[
                        ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                        ('index', models.IntegerField()),
                        ('label', models.CharField(blank=True, max_length=100, null=True)),
                        ('value', models.IntegerField()),
                        ('add_border', models.BooleanField(default=False)),
                        ('question', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm_questionnaire.questionbase')),
                    ],
                    options={
                        'ordering': ['index'],
                    },
                ),
                migrations.CreateModel(
                    name='QuestionnaireResponse',
                    fields=[
                        ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                        ('time_submitted', models.DateTimeField(default=django.utils.timezone.now)),
                        ('data', models.BinaryField()),
                        ('participant', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.participant')),
                        ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.donationproject')),
                    ],
                    options={
                        'abstract': False,
                    },
                ),
                migrations.CreateModel(
                    name='QuestionItem',
                    fields=[
                        ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                        ('label', models.CharField(blank=True, max_length=255, null=True)),
                        ('label_alt', models.CharField(blank=True, max_length=255, null=True)),
                        ('index', models.IntegerField()),
                        ('value', models.IntegerField()),
                        ('randomize', models.BooleanField(default=False)),
                        ('question', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm_questionnaire.questionbase')),
                    ],
                    options={
                        'ordering': ['index'],
                    },
                ),
                migrations.AddConstraint(
                    model_name='scalepoint',
                    constraint=models.UniqueConstraint(fields=('index', 'question'), name='unique_index_per_question'),
                ),
                migrations.AddConstraint(
                    model_name='questionitem',
                    constraint=models.UniqueConstraint(fields=('index', 'question'), name='unique_item_index_per_question'),
                ),
                migrations.AddConstraint(
                    model_name='questionitem',
                    constraint=models.UniqueConstraint(fields=('value', 'question'), name='unique_item_value_per_question'),
                ),
                migrations.AddConstraint(
                    model_name='questionbase',
                    constraint=models.UniqueConstraint(fields=('variable_name', 'project'), name='unique_varname_per_project'),
                ),
            ],
            # Table already exists. See ddm/migrations/0051_auto_20241025_1209.py
            database_operations=[],
        )
    ]



# Generated by Django 3.2.13 on 2024-10-25 08:40

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('ddm', '0049_delete_projectaccesstoken'),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.CreateModel(
                    name='ProjectAccessToken',
                    fields=[
                        ('key', models.CharField(max_length=40, primary_key=True, serialize=False)),
                        ('created', models.DateTimeField(auto_now_add=True)),
                        ('expiration_date', models.DateTimeField(blank=True, null=True)),
                        ('project', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE,
                                                         related_name='donation_project', to='ddm.donationproject',
                                                         verbose_name='Donation Project')),
                    ],
                )
            ],
            # Table already exists. See ddm/migrations/0049_delete_projectaccesstoken.py (may be moved to ddm.core)
            database_operations = [],
        )
    ]



from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.test import TestCase, override_settings

from ddm.projects.models import DonationProject, ResearchProfile


User = get_user_model()


@override_settings(DDM_SETTINGS={'EMAIL_PERMISSION_CHECK':  r'.*(\.|@)mail\.com$', })
class TestDonationProject(TestCase):
    @classmethod
    def setUpTestData(cls):
        credentials = {
            'username': 'no_prof', 'password': '123', 'email': 'u@mail.com'
        }
        cls.user = User.objects.create_user(**credentials)
        cls.user_profile = ResearchProfile.objects.create(user=cls.user)
        cls.project = DonationProject.objects.create(
            name='Base Project', slug='base', owner=cls.user_profile)

    def test_project_owner_cannot_be_null_on_create(self):
        invalid_project = DonationProject(name='test_project', slug='test')
        self.assertRaises(ValidationError, invalid_project.save)

    def test_get_statistics(self):
        statistics_keys = [
            'n_started',
            'n_completed',
            'completion_rate',
            'n_donations',
            'n_errors',
            'average_time'
        ]
        for key in statistics_keys:
            self.assertIn(key, self.project.get_statistics())

    def test_create_url_id(self):
        new_project = DonationProject.objects.create(
            name='test project', slug='test', owner=self.user_profile)
        self.assertIsNotNone(new_project.url_id)
        self.assertEqual(len(new_project.url_id), 8)



from django.views.decorators.debug import sensitive_variables
from rest_framework.fields import empty


class SerializerDecryptionMixin:
    """
    Allows to pass the secret for the decryption of super secret projects to
    the serializer on init.  # TODO: Improve description.
    """
    @sensitive_variables()
    def __init__(self, instance=None, data=empty, decryptor=None, **kwargs):
        self.decryptor = decryptor
        self.secret = kwargs.pop('secret', None)
        super().__init__(instance=instance, data=data, **kwargs)

    @sensitive_variables()
    def get_data(self, obj):
        if not self.secret:
            self.secret = obj.project.secret_key
        return obj.get_decrypted_data(self.secret, obj.project.get_salt(), self.decryptor)



from django.contrib.auth import get_user_model
from django.contrib.messages.views import SuccessMessageMixin
from django.http import Http404
from django.shortcuts import reverse, redirect
from django.views.generic import FormView
from django.views.generic.base import TemplateView

from ddm.auth.forms import TokenCreationForm
from ddm.auth.models import ProjectAccessToken
from ddm.auth.utils import user_is_permitted, user_has_project_access
from ddm.projects.models import DonationProject, ResearchProfile


User = get_user_model()


class DDMAuthMixin:
    """
    Mixin for Class Based Views that handles redirects as follows:
    - unauthenticated users => login page.
    - users without research profiles => create research profile.
    - users without authorization (e.g., due to e-mail restriction) => no permission page.
    - users without owner-rights => 404.
    """
    def dispatch(self, request, *args, **kwargs):
        if request.method in ['GET', 'POST']:
            if not request.user.is_authenticated:
                return redirect('ddm_login')

            if not ResearchProfile.objects.filter(user=request.user).exists():
                ResearchProfile.objects.create(user=request.user)

            if not user_is_permitted(request.user):
                return redirect('ddm_auth:no_permission')

            if request.path not in [reverse('ddm_projects:list'), reverse('ddm_projects:create')]:
                if 'project_url_id' in self.kwargs:
                    project_id = self.kwargs['project_url_id']
                else:
                    raise Http404()

                try:
                    project = DonationProject.objects.get(url_id=project_id)
                except DonationProject.DoesNotExist:
                    raise Http404()

                if (not user_has_project_access(request.user, project) and
                        not request.user.is_superuser):
                    raise Http404()
        return super().dispatch(request, *args, **kwargs)


class DdmNoPermissionView(TemplateView):
    """
    View to inform users that they do not have the needed permission rights.
    Implements the following redirects:
    * Unauthenticated users are redirected to the login page.
    * Logged-in users with permission and a research profile are redirected to the project list.
    """
    template_name = 'ddm_auth/no_permission.html'

    def dispatch(self, request, *args, **kwargs):
        if request.method == 'GET':
            if not request.user.is_authenticated:
                return redirect('ddm_login')
            elif (user_is_permitted(request.user) and
                  ResearchProfile.objects.filter(user=request.user).exists()):
                return redirect('ddm_projects:list')
        return super().dispatch(request, *args, **kwargs)


class ProjectTokenView(SuccessMessageMixin, DDMAuthMixin, FormView):
    """ View to see existing access token or generate a new one. """
    template_name = 'ddm_auth/token.html'
    form_class = TokenCreationForm

    def get_project(self):
        """ Returns current project. """
        project_url_id = self.kwargs.get('project_url_id')
        return DonationProject.objects.filter(url_id=project_url_id).first()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        project = self.get_project()
        context.update({
            'project': project,
            'token': ProjectAccessToken.objects.filter(project=project).first()
        })
        return context

    def form_valid(self, form):
        """ If form is valid, either create new token or delete existing one. """
        project = self.get_project()
        if form.cleaned_data['action'] == 'create':
            project.create_token(expiration_days=form.cleaned_data['expiration_days'])

        if form.cleaned_data['action'] == 'delete':
            token = project.get_token()
            token.delete()

        return super().form_valid(form)

    def get_success_url(self):
        return reverse('ddm_auth:project_token', kwargs={'project_url_id': self.kwargs.get('project_url_id')})



import re

from django.conf import settings

from ddm.projects.models import DonationProject, ResearchProfile


def email_is_valid(email_string):
    """
    Check if an email address complies with the pattern specified in
    DDM_SETTINGS['EMAIL_PERMISSION_CHECK'].
    If this setting is not defined, returns True.
    """
    if hasattr(settings, 'DDM_SETTINGS'):
        if 'EMAIL_PERMISSION_CHECK' in settings.DDM_SETTINGS:
            match = re.match(settings.DDM_SETTINGS['EMAIL_PERMISSION_CHECK'],
                             email_string)
            if not match:
                return False
    return True


def user_has_project_access(user, project):
    """
    Returns true if user is owner or collaborator of a project. False otherwise.
    """
    if user.is_anonymous:
        return False

    user_profile = ResearchProfile.objects.filter(user=user).first()
    if not user_profile:
        return False
    else:
        if project.owner == user_profile:
            return True
        else:
            return False


def user_is_permitted(user):
    """
    Check if a user has access permission.
    """
    if user.is_superuser:
        return True
    elif user.is_authenticated:
        profile = ResearchProfile.objects.filter(user=user).first()
        if profile:
            if profile.ignore_email_restriction:
                return True
        if email_is_valid(user.email):
            return True
    return False



class QuestionValidationError(Exception):
    def __init__(self, question, message=None, item_errors=None, scale_errors=None):
        if item_errors is None:
            item_errors = []
        if scale_errors is None:
            scale_errors = []

        question_type = question.DEFAULT_QUESTION_TYPE

        item_error_prefix = f'Item error for {question_type} {question.pk}: '
        self.item_errors = [item_error_prefix + str(e) for e in item_errors]

        scale_error_prefix = f'Scale error for {question_type} {question.pk}: '
        self.scale_errors = [scale_error_prefix + str(e) for e in scale_errors]

        error_message = f'Validation error for {question_type} {question.pk}: {message}.'

        errors = self.item_errors + self.scale_errors
        if message is not None:
            errors.append(error_message)
        self.errors = errors

        # check if message is provided and add to errors.
        super().__init__(error_message)



# Generated by Django 4.2.16 on 2024-12-04 15:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm_projects', '0002_alter_donationproject_briefing_text_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='donationproject',
            name='active',
            field=models.BooleanField(default=True, help_text='Participants can only take part in a project if it is active.', verbose_name='Active'),
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='redirect_target',
            field=models.CharField(blank=True, help_text='Always include <i>http://</i> or <i>https://</i> in the redirect address. If URL parameter extraction is enabled for this project, you can include the extracted URL parameters in the redirect address as follows: "https://redirect.me/?redirectpara=<b>{{participant.data.url_param.URLParameter}}</b>".', max_length=2000, verbose_name='Redirect address'),
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='slug',
            field=models.SlugField(help_text='Identifier that is included in the URL through which participants can access the project (e.g, https://root.url/url-identifier). Can only contain letters, hyphens, numbers or underscores.', unique=True, verbose_name='URL Identifier'),
        ),
    ]



from django import template


register = template.Library()

# Register template tags.
register.tag('for', template.defaulttags.do_for)
register.tag('if', template.defaulttags.do_if)
register.tag('ifchanged', template.defaulttags.ifchanged)
register.tag('regroup', template.defaulttags.regroup)

# Register template filters.
register.filter('date', template.defaultfilters.date, expects_localtime=True, is_safe=False)
register.filter('default', template.defaultfilters.default, is_safe=False)
register.filter('dictsort', template.defaultfilters.dictsort, is_safe=False)
register.filter('dictsortreversed', template.defaultfilters.dictsortreversed, is_safe=False)
register.filter('first', template.defaultfilters.first, is_safe=False)
register.filter('last', template.defaultfilters.last, is_safe=True)
register.filter('length', template.defaultfilters.length, is_safe=False)
register.filter('random', template.defaultfilters.random, is_safe=True)
register.filter('truncatechars', template.defaultfilters.truncatechars, is_safe=True)
register.filter('truncatewords', template.defaultfilters.truncatewords, is_safe=True)



from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.test import TestCase
from django.utils import timezone

from ddm.datadonation.models import DonationInstruction, FileUploader, DonationBlueprint, DataDonation, ProcessingRule
from ddm.logging.models import ExceptionLogEntry
from ddm.participation.models import Participant
from ddm.projects.models import DonationProject, ResearchProfile


User = get_user_model()


class TestDonationInstructionModel(TestCase):
    @classmethod
    def setUpTestData(cls):
        user = User.objects.create_user(**{
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        })
        profile = ResearchProfile.objects.create(user=user)

        project = DonationProject.objects.create(
            name='Base Project', slug='base', owner=profile)

        cls.file_uploader = FileUploader.objects.create(
            project=project,
            name='basic file uploader',
            upload_type=FileUploader.UploadTypes.SINGLE_FILE
        )

    def setUp(self):
        self.inst_2 = DonationInstruction.objects.create(
            text='instruction 2',
            index=2,
            file_uploader=self.file_uploader
        )
        self.inst_3 = DonationInstruction.objects.create(
            text='instruction 3',
            index=3,
            file_uploader=self.file_uploader
        )

    def test_create_with_existing_index_pushes_other_indices(self):
        new_instruction = DonationInstruction.objects.create(
            text='instruction new',
            index=2,
            file_uploader=self.file_uploader
        )
        self.assertEqual(new_instruction.index, 2)

        self.inst_2.refresh_from_db()
        self.assertEqual(self.inst_2.index, 3)

        self.inst_3.refresh_from_db()
        self.assertEqual(self.inst_3.index, 4)

    def test_decrease_index_of_existing_instruction(self):
        self.inst_3.index = 2
        self.inst_3.save()
        self.assertEqual(self.inst_3.index, 2)

        self.inst_2.refresh_from_db()
        self.assertEqual(self.inst_2.index, 3)

    def test_increase_index_of_existing_instruction(self):
        self.inst_2.index = 3
        self.inst_2.save()
        self.assertEqual(self.inst_2.index, 3)

        self.inst_3.refresh_from_db()
        self.assertEqual(self.inst_3.index, 2)

    def test_indices_are_adjusted_on_delete(self):
        self.inst_2.delete()

        self.inst_3.refresh_from_db()
        self.assertEqual(self.inst_3.index, 2)

    def test_clean(self):
        instruction_in_db = DonationInstruction.objects.create(
            text='instruction',
            index=20,
            file_uploader=self.file_uploader
        )
        instruction_not_in_db = DonationInstruction(
            text='instruction',
            index=21,
            file_uploader=self.file_uploader
        )
        with self.assertRaises(ValidationError):
            instruction_in_db.clean()
            instruction_not_in_db.clean()


class TestDonationBlueprintModel(TestCase):
    @classmethod
    def setUpTestData(cls):
        user = User.objects.create_user(**{
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        })
        profile = ResearchProfile.objects.create(user=user)

        project = DonationProject.objects.create(
            name='Base Project', slug='base', owner=profile)

        cls.file_uploader = FileUploader.objects.create(
            project=project,
            name='basic file uploader',
            upload_type=FileUploader.UploadTypes.SINGLE_FILE
        )

        cls.blueprint = DonationBlueprint.objects.create(
            project=project,
            name='valid blueprint',
            description='some description',
            expected_fields='"some field"',
            file_uploader=cls.file_uploader,
            regex_path='/this/file.json'
        )

        cls.participant = Participant.objects.create(
            project=project,
            start_time=timezone.now()
        )

        ProcessingRule.objects.create(
            blueprint=cls.blueprint,
            name='',
            field='fieldA',
            execution_order=1,
        )

        ProcessingRule.objects.create(
            blueprint=cls.blueprint,
            name='',
            field='fieldB',
            execution_order=2,
            comparison_operator=ProcessingRule.ComparisonOperators.EQUAL,
        )

        ProcessingRule.objects.create(
            blueprint=cls.blueprint,
            name='',
            field='fieldC',
            execution_order=3,
        )

    def test_get_slug(self):
        self.assertEqual(self.blueprint.get_slug(), 'blueprint')

    def test_validate_donation_case_valid(self):
        data = {
            'consent': True,
            'extracted_data': ['some data'],
            'status': 'success'
        }
        self.assertTrue(self.blueprint.validate_donation(data))

    def test_validate_donation_case_invalid(self):
        data = {
            'consent': True,
            'extracted_data': ['some data'],
        }
        self.assertFalse(self.blueprint.validate_donation(data))

    def test_process_donation_case_valid(self):
        data = {
            'consent': True,
            'extracted_data': ['some data'],
            'status': 'success'
        }
        n_donations_pre = DataDonation.objects.count()
        self.blueprint.process_donation(data, self.participant)
        n_donations_post = DataDonation.objects.count()
        self.assertEqual(n_donations_post - n_donations_pre, 1)

    def test_process_donation_case_invalid(self):
        data = {
            'consent': True,
            'extracted_data': ['some data'],
        }
        n_donations_pre = DataDonation.objects.count()
        n_exceptions_pre = ExceptionLogEntry.objects.count()

        self.blueprint.process_donation(data, self.participant)

        n_donations_post = DataDonation.objects.count()
        n_exceptions_post = ExceptionLogEntry.objects.count()

        self.assertEqual(n_donations_post - n_donations_pre, 0)
        self.assertEqual(n_exceptions_post - n_exceptions_pre, 2)

    def test_get_fields_to_extract(self):
        expected_fields = ['fieldA', 'fieldC']
        self.assertEqual(
            sorted(self.blueprint.get_fields_to_extract()), sorted(expected_fields))


class TestProcessingRuleModel(TestCase):
    @classmethod
    def setUpTestData(cls):
        user = User.objects.create_user(**{
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        })
        profile = ResearchProfile.objects.create(user=user)

        project = DonationProject.objects.create(
            name='Base Project', slug='base', owner=profile)

        cls.file_uploader = FileUploader.objects.create(
            project=project,
            name='basic file uploader',
            upload_type=FileUploader.UploadTypes.SINGLE_FILE
        )

        cls.blueprint = DonationBlueprint.objects.create(
            project=project,
            name='valid blueprint',
            description='some description',
            expected_fields='"some field"',
            file_uploader=cls.file_uploader,
            regex_path='/this/file.json'
        )

        cls.participant = Participant.objects.create(
            project=project,
            start_time=timezone.now()
        )

    def test_get_rule_config(self):
        rule = ProcessingRule.objects.create(
            blueprint=self.blueprint,
            name='test name',
            field='some field',
            execution_order=1,
            comparison_operator=ProcessingRule.ComparisonOperators.EMPTY,
        )
        expected_config = {
            'field': 'some field',
            'regex_field': False,
            'comparison_operator': ProcessingRule.ComparisonOperators.EMPTY,
            'comparison_value': '',
            'replacement_value': ''
        }
        self.assertEqual(rule.get_rule_config(), expected_config)



import io
import json
import zipfile

from django import forms
from django.contrib import messages
from django.contrib.messages.views import SuccessMessageMixin
from django.db.models import Q
from django.http import HttpResponseRedirect, HttpResponse, Http404
from django.shortcuts import get_object_or_404
from django.urls import reverse, reverse_lazy
from django.utils import timezone
from django.views.decorators.debug import sensitive_variables
from django.views.generic.edit import CreateView, UpdateView, DeleteView, FormView
from django.views.generic.list import ListView

from ddm.apis.serializers import DataDonationSerializer
from ddm.apis.views import DDMAPIMixin
from ddm.datadonation.forms import (
    BlueprintEditForm, ProcessingRuleInlineFormset, SecretInputForm, InstructionsForm
)
from ddm.datadonation.models import (
    DonationBlueprint, DonationInstruction, FileUploader
)
from ddm.encryption.models import Decryption, Encryption
from ddm.participation.models import Participant
from ddm.projects.models import DonationProject
from ddm.projects.views import DDMAuthMixin


class BlueprintMixin:
    """ Mixin for all blueprint related views. """
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        file_uploaders = FileUploader.objects.filter(project__url_id=self.kwargs['project_url_id'])
        context.update({
            'project': DonationProject.objects.get(url_id=self.kwargs['project_url_id']),
            'file_uploader_meta': {str(fu.pk): fu.upload_type for fu in file_uploaders}
        })
        return context

    def get_success_url(self):
        return reverse(
            'ddm_datadonation:overview',
            kwargs={'project_url_id': self.kwargs['project_url_id']}
        )


class DataDonationOverview(DDMAuthMixin, BlueprintMixin, ListView):
    """ View to list all file uploaders associated with a project. """
    model = FileUploader
    context_object_name = 'file_uploaders'
    template_name = 'ddm_datadonation/overview.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({'lonely_blueprints': context['project'].donationblueprint_set.filter(file_uploader=None)})
        return context

    def get_queryset(self):
        queryset = super().get_queryset().filter(project__url_id=self.kwargs['project_url_id'])
        return queryset


class FileUploaderCreate(SuccessMessageMixin, DDMAuthMixin, BlueprintMixin, CreateView):
    """ View to create a new file uploader. """
    model = FileUploader
    template_name = 'ddm_datadonation/uploader/create.html'
    fields = ['name', 'upload_type', 'combined_consent']
    success_message = 'File Uploader was created successfully.'

    def form_valid(self, form):
        project_url_id = self.kwargs['project_url_id']
        project = DonationProject.objects.get(url_id=project_url_id)
        form.instance.project_id = project.pk
        return super().form_valid(form)

    def get_success_url(self):
        kwargs = {
            'project_url_id': self.kwargs['project_url_id'],
            'pk': self.object.pk
        }
        return reverse('ddm_datadonation:uploaders:edit', kwargs=kwargs)


class FileUploaderEdit(SuccessMessageMixin, DDMAuthMixin, BlueprintMixin, UpdateView):
    """ View to edit the details of an existing file uploader. """
    model = FileUploader
    template_name = 'ddm_datadonation/uploader/edit.html'
    fields = ['name', 'upload_type', 'combined_consent', 'index']
    success_message = 'Blueprint Uploader "%(name)s" was successfully updated.'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({'blueprints': self.get_relevant_blueprints()})
        return context

    def get_relevant_blueprints(self):
        """
        Returns a query set containing all blueprints that are associated with
        the current or no file uploader.
        """
        relevant_blueprints = DonationBlueprint.objects.filter(
            Q(file_uploader=self.object) | Q(file_uploader=None), project__url_id=self.kwargs['project_url_id'])
        return relevant_blueprints

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        form = self.get_form()
        selected_blueprints = [int(k[3:]) for k in self.request.POST.keys() if k.startswith('bp-')]
        if form.is_valid():
            return self.form_valid(form, selected_blueprints)
        else:
            return self.form_invalid(form)

    def form_valid(self, form, selected_blueprints):
        self.object = form.save()

        # Update file_upload foreign key on donation blueprints.
        relevant_blueprints = self.get_relevant_blueprints()
        for bp in relevant_blueprints:
            if bp.pk in selected_blueprints:
                bp.file_uploader = self.object
                bp.save()
            elif bp.file_uploader == self.object:
                bp.file_uploader = None
                bp.save()
        messages.add_message(
            self.request, messages.SUCCESS,
            self.success_message % dict(name=self.object.name),
            fail_silently=True,
        )
        return HttpResponseRedirect(self.get_success_url())


class FileUploaderDelete(SuccessMessageMixin, DDMAuthMixin, BlueprintMixin, DeleteView):
    """ View to delete an existing blueprint uploader. """
    model = FileUploader
    template_name = 'ddm_datadonation/uploader/delete.html'
    success_message = 'File Uploader "%s" was deleted.'

    def get_success_message(self, cleaned_data):
        return self.success_message % self.object.name


class BlueprintCreate(SuccessMessageMixin, DDMAuthMixin, BlueprintMixin, CreateView):
    """ View to create a new donation blueprint. """
    model = DonationBlueprint
    template_name = 'ddm_datadonation/blueprint/create.html'
    form_class = BlueprintEditForm
    success_message = 'Blueprint was created successfully.'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        available_file_uploaders = FileUploader.objects.filter(project__url_id=self.kwargs['project_url_id'])
        context['form'].fields['file_uploader'].queryset = available_file_uploaders
        return context

    def form_valid(self, form):
        project_url_id = self.kwargs['project_url_id']
        project = DonationProject.objects.get(url_id=project_url_id)
        form.instance.project_id = project.pk
        return super().form_valid(form)

    def get_success_url(self):
        return reverse('ddm_datadonation:blueprints:edit',
                       kwargs={'project_url_id': self.object.project.url_id, 'pk': self.object.pk})


class BlueprintEdit(SuccessMessageMixin, DDMAuthMixin, BlueprintMixin, UpdateView):
    """ View to edit the details of an existing donation blueprint. """
    model = DonationBlueprint
    template_name = 'ddm_datadonation/blueprint/edit.html'
    form_class = BlueprintEditForm
    success_message = 'Blueprint "%(name)s" was successfully updated.'

    def get_success_url(self):
        return reverse(
            'ddm_datadonation:overview',
            kwargs={'project_url_id': self.object.project.url_id}
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        available_file_uploaders = FileUploader.objects.filter(
            project__url_id=self.kwargs['project_url_id'])
        context['form'].fields['file_uploader'].queryset = available_file_uploaders
        context['formset'] = ProcessingRuleInlineFormset(
            instance=self.object, queryset=self.object.processingrule_set.order_by('execution_order'))
        return context

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        form = self.get_form()
        formset = ProcessingRuleInlineFormset(self.request.POST, instance=self.object)
        if form.is_valid() and formset.is_valid():
            return self.form_valid(form, formset)
        else:
            return self.form_invalid(form, formset)

    def form_valid(self, form, formset):
        self.object = form.save()
        formset.instance = self.object
        formset.save()
        messages.add_message(
            self.request, messages.SUCCESS,
            self.success_message % dict(name=self.object.name),
            fail_silently=True,
        )
        return HttpResponseRedirect(self.get_success_url())

    def form_invalid(self, form, formset):
        return self.render_to_response(self.get_context_data(form=form, formset=formset))


class BlueprintDelete(SuccessMessageMixin, DDMAuthMixin, BlueprintMixin, DeleteView):
    """ View to delete an existing donation blueprint. """
    model = DonationBlueprint
    template_name = 'ddm_datadonation/blueprint/delete.html'
    success_message = 'Blueprint "%s" was deleted.'

    def get_success_message(self, cleaned_data):
        return self.success_message % self.object.name


class InstructionMixin:
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({'project_url_id': self.kwargs['project_url_id']})
        context.update({'file_uploader': FileUploader.objects.get(pk=self.kwargs['file_uploader_pk'])})
        return context

    def get_success_url(self):
        kwargs = {'project_url_id': self.kwargs['project_url_id'],
                  'file_uploader_pk': self.kwargs['file_uploader_pk']}
        return reverse('ddm_datadonation:instructions:overview', kwargs=kwargs)


class InstructionOverview(DDMAuthMixin, InstructionMixin, ListView):
    """ View to create a new instruction page. """
    model = DonationInstruction
    context_object_name = 'instructions'
    template_name = 'ddm_datadonation/instructions/list.html'
    fields = ['text', 'index']

    def get_queryset(self):
        queryset = super().get_queryset().filter(file_uploader_id=self.kwargs['file_uploader_pk'])
        return queryset


class InstructionCreate(SuccessMessageMixin, DDMAuthMixin, InstructionMixin, CreateView):
    """ View to create an instruction page. """
    model = DonationInstruction
    form_class = InstructionsForm
    template_name = 'ddm_datadonation/instructions/create.html'
    success_message = 'Instruction page was successfully created.'

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['instance'] = DonationInstruction(file_uploader=FileUploader.objects.get(id=self.kwargs['file_uploader_pk']))
        return kwargs

    def get_initial(self):
        initial = super().get_initial()
        related_file_uploader = FileUploader.objects.get(id=self.kwargs['file_uploader_pk'])
        indices = related_file_uploader.donationinstruction_set.values_list('index', flat=True)
        if indices:
            initial['index'] = max(indices) + 1
        else:
            initial['index'] = 1
        return initial


class InstructionEdit(SuccessMessageMixin, DDMAuthMixin, InstructionMixin, UpdateView):
    """ View to edit an instruction page. """
    model = DonationInstruction
    form_class = InstructionsForm
    template_name = 'ddm_datadonation/instructions/edit.html'
    success_message = 'Instruction page was successfully updated.'


class InstructionDelete(SuccessMessageMixin, DDMAuthMixin, InstructionMixin, DeleteView):
    """ View to delete an instruction page. """
    model = DonationInstruction
    template_name = 'ddm_datadonation/instructions/delete.html'
    success_message = 'Instruction page was deleted.'

    def get_success_message(self, cleaned_data):
        return self.success_message % self.object.name


class DonationDownloadView(DDMAuthMixin, DDMAPIMixin, FormView):
    """View to download all the donations of one specific participant."""

    def dispatch(self, request, *args, **kwargs):
        try:
            self.project = self.get_project()
        except Http404 as e:
            raise Http404(str(e))
        return super().dispatch(request, *args, **kwargs)

    def get_form_class(self):
        return SecretInputForm if self.project.super_secret else forms.Form

    def get_template_names(self):
        regular_template = 'ddm_datadonation/download_regular.html'
        secret_template = 'ddm_datadonation/download_with_secret.html'
        return secret_template if self.project.super_secret else regular_template

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({
            'project': self.project,
            'participant_id': self.kwargs.get('participant_id'),
        })
        return context

    def get_success_url(self):
        url = reverse_lazy(
            'ddm_datadonation:download_donation',
            kwargs={
                'project_url_id': self.kwargs.get('project_url_id'),
                'participant_id': self.kwargs.get('participant_id')
            })
        return url

    def get_participant(self):
        """ Returns participant instance. """
        participant_id = self.kwargs.get('participant_id')
        return get_object_or_404(Participant, external_id=participant_id)

    def get_donations(self, participant):
        return participant.datadonation_set.all()

    @sensitive_variables()
    def form_valid(self, form):
        try:
            decryptor = self.get_decryptor(form.cleaned_data.get('secret'))
        except ValueError as e:
            form.add_error('secret', 'Incorrect secret.')
            self.create_event_log(
                descr='Donation download failed.',
                msg=f'Download attempt with incorrect secret was registered.'
            )
            return HttpResponse('Incorrect secret', status=422)

        return self.get_data_response(decryptor)

    @sensitive_variables()
    def get_decryptor(self, secret=None):
        secret_key = secret if secret else self.project.secret_key
        decryptor = Decryption(secret_key, self.project.get_salt())
        try:
            self.test_secret(secret_key, decryptor)
        except ValueError:
            raise ValueError
        return decryptor

    @sensitive_variables()
    def test_secret(self, secret, decryptor):
        test_string = 'Teststring'
        encrypted_string = Encryption(public=self.project.public_key).encrypt(test_string)

        try:
            decrypted_string = decryptor.decrypt(encrypted_string)
        except ValueError:
            raise ValueError

        if decrypted_string != test_string:
            raise ValueError

    def get_filename(self, participant_id, file_type):
        return f'ddm_{self.project.url_id}_donations_{participant_id}.{file_type}'

    def create_zip(self, content, participant_id):
        """ Creates a zip file in memory. """
        buffer = io.BytesIO()
        filename = self.get_filename(participant_id, 'json')
        with zipfile.ZipFile(buffer, mode='w', compression=zipfile.ZIP_DEFLATED) as zf:
            with zf.open(filename, 'w') as json_file:
                json_content = json.dumps(
                    content, ensure_ascii=False, separators=(',', ':'))
                json_file.write(json_content.encode('utf-8'))
                zf.testzip()
        zip_in_memory = buffer.getvalue()
        buffer.flush()
        return zip_in_memory

    def create_zip_response(self, zip_file, participant_id):
        """ Creates an HttpResponse object containing the provided zip file. """
        filename = self.get_filename(participant_id, 'zip')
        response = HttpResponse(zip_file, content_type='application/zip')
        response['Content-Length'] = len(zip_file)
        response['Content-Disposition'] = f'attachment; filename={filename}.zip'
        return response

    @sensitive_variables()
    def get_data_response(self, decryptor, *args, **kwargs):
        participant = self.get_participant()
        donations = self.get_donations(participant)

        result = []
        for donation in donations:
            result.append({
                'blueprint_id': donation.blueprint.pk,
                'blueprint_name': donation.blueprint.name,
                'donation': DataDonationSerializer(donation, decryptor=decryptor).data,
            })

        zip_in_mem = self.create_zip(result, participant.external_id)
        response = self.create_zip_response(zip_in_mem, participant.external_id)

        self.create_event_log(
            descr='Donation download registered.',
            msg=f'Donation for participant {participant} downloaded.'
        )
        return response



from django.contrib.messages.views import SuccessMessageMixin
from django.views.generic import TemplateView

from ddm.auth.views import DDMAuthMixin
from ddm.projects.models import DonationProject


class ProjectLogsView(SuccessMessageMixin, DDMAuthMixin, TemplateView):
    """ View that lists all exceptions related to a project. """
    template_name = 'ddm_logging/overview.html'

    def get_project(self):
        project_url_id = self.kwargs.get('project_url_id')
        return DonationProject.objects.filter(url_id=project_url_id).first()

    def get_event_logs(self):
        project = self.get_project()
        return project.eventlogentry_set.all()

    def get_exception_logs(self):
        project = self.get_project()
        return project.exceptionlogentry_set.all()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({
            'project': self.get_project(),
            'events': self.get_event_logs(),
            'exceptions': self.get_exception_logs()
        })
        return context



from django.contrib.auth import get_user_model
from django.test import TestCase, Client
from django.urls import reverse

from ddm.datadonation.models import FileUploader, DonationBlueprint
from ddm.logging.models import ExceptionLogEntry, ExceptionRaisers
from ddm.projects.models import DonationProject, ResearchProfile


User = get_user_model()


class TestExceptionAPI(TestCase):
    @classmethod
    def setUpTestData(cls):
        user = User.objects.create_user(**{
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        })
        profile = ResearchProfile.objects.create(user=user)

        cls.project = DonationProject.objects.create(
            name='Base Project', slug='base', owner=profile)

        cls.file_uploader = FileUploader.objects.create(
            project=cls.project,
            name='zip file uploader',
            upload_type=FileUploader.UploadTypes.ZIP_FILE
        )

        cls.blueprint = DonationBlueprint.objects.create(
            project=cls.project,
            name='valid blueprint',
            description='some description',
            expected_fields='"some field"',
            file_uploader=cls.file_uploader,
            regex_path='/this/file.json'
        )

        cls.post_url = reverse('ddm_logging:exceptions_api', args=[cls.project.url_id])
        cls.post_data = {
            'blueprint': cls.blueprint.pk,
            'status_code': 1,
            'raised_by': ExceptionRaisers.SERVER,
            'message': 'Some message.'
        }

    def test_post_without_participant(self):
        client = Client()
        exceptions_count_before = ExceptionLogEntry.objects.count()
        client.post(self.post_url, self.post_data)
        exceptions_count_after = ExceptionLogEntry.objects.count()
        self.assertEqual(exceptions_count_before, (exceptions_count_after - 1))

    def test_valid_post(self):
        client = Client()
        client.get(reverse('ddm_participation:briefing', args=[self.project.slug]))
        exceptions_count_before = ExceptionLogEntry.objects.count()
        client.post(self.post_url, self.post_data)
        exceptions_count_after = ExceptionLogEntry.objects.count()
        self.assertEqual(exceptions_count_before, (exceptions_count_after - 1))



# Generated by Django 3.2.13 on 2024-10-25 09:19

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('ddm', '0050_auto_20241025_1119'),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.CreateModel(
                    name='ExceptionLogEntry',
                    fields=[
                        ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                        ('date', models.DateTimeField(default=django.utils.timezone.now)),
                        ('raised_by', models.CharField(blank=True, choices=[('server', 'Server'), ('client', 'Client')], max_length=20)),
                        ('exception_type', models.IntegerField(null=True)),
                        ('message', models.TextField()),
                        ('blueprint', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='ddm.donationblueprint')),
                        ('participant', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='ddm.participant')),
                        ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.donationproject')),
                    ],
                ),
                migrations.CreateModel(
                    name='EventLogEntry',
                    fields=[
                        ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                        ('date', models.DateTimeField(default=django.utils.timezone.now)),
                        ('description', models.TextField()),
                        ('message', models.TextField()),
                        ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.donationproject')),
                    ],
                ),
            ],
            # Table already exists. See ddm/migrations/0050_auto_20241025_1119.py (may be moved to ddm.core)
            database_operations=[],
        )
    ]



# Generated by Django 4.2.16 on 2024-11-07 10:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm_datadonation', '0003_auto_20241026_1048'),
    ]

    operations = [
        migrations.AlterField(
            model_name='donationinstruction',
            name='text',
            field=models.TextField(default=''),
            preserve_default=False,
        ),
    ]



from django.contrib.auth import get_user_model
from django.test import TestCase, override_settings, Client
from django.urls import reverse
from django.utils import timezone

from ddm.datadonation.models import DonationBlueprint, FileUploader, DataDonation
from ddm.participation.models import Participant

from ddm.projects.models import ResearchProfile, DonationProject

User = get_user_model()


@override_settings(DDM_SETTINGS={'EMAIL_PERMISSION_CHECK':  r'.*(\.|@)mail\.com$', })
class BlueprintEditTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = User.objects.create_user(**{
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        })
        profile = ResearchProfile.objects.create(user=cls.user)

        cls.project = DonationProject.objects.create(
            name='Base Project', slug='base', owner=profile)

        cls.file_uploader = FileUploader.objects.create(
            project=cls.project,
            name='basic file uploader',
            upload_type=FileUploader.UploadTypes.ZIP_FILE
        )

        cls.blueprint = DonationBlueprint.objects.create(
            project=cls.project,
            name='valid blueprint',
            description='some description',
            expected_fields='"some field"',
            file_uploader=cls.file_uploader,
            regex_path='/this/file.json'
        )

        cls.url = reverse('ddm_datadonation:blueprints:edit',
                          kwargs={'pk': cls.blueprint.pk, 'project_url_id': cls.project.url_id})

    def test_post_valid_data(self):
        valid_data = {
            'name': 'some name',
            'description': 'some description',
            'regex_path': '/file.path',
            'exp_file_format': DonationBlueprint.FileFormats.JSON_FORMAT,
            'csv_delimiter': '',
            'file_uploader': self.file_uploader.pk,
            'json_extraction_root': '',
            'expected_fields': '"fieldA"',
            'expected_fields_regex_matching': 'asdflkjklsadjf',
        }
        formset_data = {
            'processingrule_set-TOTAL_FORMS': '1',
            'processingrule_set-INITIAL_FORMS': '0',
        }
        data = {**valid_data, **formset_data}
        bp_name_before = self.blueprint.name
        self.client.login(**{'username': 'owner', 'password': '123'})
        response = self.client.post(self.url, data)
        bp_name_after = DonationBlueprint.objects.get(pk=self.blueprint.pk).name

        self.assertEqual(response.status_code, 302)
        redirect_url = reverse(
            'ddm_datadonation:overview',
            kwargs={'project_url_id': self.project.url_id}
        )
        self.assertEqual(response.url, redirect_url)
        self.assertNotEqual(bp_name_before, bp_name_after)

    def test_post_invalid_data(self):
        invalid_data = {
            'name': 'some other name',
            'description': 'some other description',
            'regex_path': '',
            'exp_file_format': DonationBlueprint.FileFormats.JSON_FORMAT,
            'csv_delimiter': '1234567891011'
        }
        formset_data = {
            'processingrule_set-TOTAL_FORMS': '0',
        }
        data = {**invalid_data, **formset_data}
        bp_name_before = self.blueprint.name

        self.client.login(**{'username': 'owner', 'password': '123'})
        response = self.client.post(self.url, data)
        bp_name_after = DonationBlueprint.objects.get(pk=self.blueprint.pk).name

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'This field is required.')
        self.assertEqual(bp_name_before, bp_name_after)


@override_settings(DDM_SETTINGS={'EMAIL_PERMISSION_CHECK':  r'.*(\.|@)mail\.com$', })
class FileUploaderEditTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = User.objects.create_user(**{
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        })
        profile = ResearchProfile.objects.create(user=cls.user)

        cls.project = DonationProject.objects.create(
            name='Base Project', slug='base', owner=profile)

        cls.file_uploader = FileUploader.objects.create(
            project=cls.project,
            name='basic file uploader',
            upload_type=FileUploader.UploadTypes.ZIP_FILE
        )

        cls.blueprint_a = DonationBlueprint.objects.create(
            project=cls.project,
            name='valid blueprint',
            description='some description',
            expected_fields='"some field"',
            file_uploader=cls.file_uploader,
            regex_path='/this/file.json'
        )
        cls.blueprint_b = DonationBlueprint.objects.create(
            project=cls.project,
            name='valid blueprint',
            description='some description',
            expected_fields='"some field"',
            file_uploader=None,
            regex_path='/this/file.json'
        )

        cls.url = reverse(
            'ddm_datadonation:uploaders:edit',
            kwargs={'pk': cls.file_uploader.pk, 'project_url_id': cls.project.url_id}
        )

    def test_post_valid_data(self):
        valid_data = {
            'name': 'some name',
            'index': 1,
            'upload_type': FileUploader.UploadTypes.ZIP_FILE,
            'combined_consent': False,
            f'bp-{self.blueprint_b.pk}': True
        }
        bp_a_uploader_before = DonationBlueprint.objects.get(pk=self.blueprint_a.pk).file_uploader
        bp_b_uploader_before = DonationBlueprint.objects.get(pk=self.blueprint_b.pk).file_uploader
        self.client.login(**{'username': 'owner', 'password': '123'})
        response = self.client.post(self.url, valid_data)
        bp_a_uploader_after = DonationBlueprint.objects.get(pk=self.blueprint_a.pk).file_uploader
        bp_b_uploader_after = DonationBlueprint.objects.get(pk=self.blueprint_b.pk).file_uploader

        self.assertEqual(response.status_code, 302)
        redirect_url = reverse(
            'ddm_datadonation:overview',
            kwargs={'project_url_id': self.project.url_id}
        )
        self.assertEqual(
            response.url,
            redirect_url
        )
        self.assertEqual(
            FileUploader.objects.get(pk=self.file_uploader.pk).name,
            valid_data['name']
        )
        self.assertNotEqual(
            bp_a_uploader_before,
            bp_a_uploader_after
        )
        self.assertNotEqual(
            bp_b_uploader_before,
            bp_b_uploader_after
        )

    def test_post_invalid_data(self):
        invalid_data = {
            'name': 'some other name',
            'upload_type': FileUploader.UploadTypes.ZIP_FILE,
            'combined_consent': False
        }
        bp_a_uploader_before = DonationBlueprint.objects.get(pk=self.blueprint_a.pk).file_uploader
        bp_b_uploader_before = DonationBlueprint.objects.get(pk=self.blueprint_b.pk).file_uploader
        self.client.login(**{'username': 'owner', 'password': '123'})
        response = self.client.post(self.url, invalid_data)
        bp_a_uploader_after = DonationBlueprint.objects.get(pk=self.blueprint_a.pk).file_uploader
        bp_b_uploader_after = DonationBlueprint.objects.get(pk=self.blueprint_b.pk).file_uploader

        # Check that the response renders the form with errors (form submission unsuccessful)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'This field is required.')
        self.assertNotEqual(
            FileUploader.objects.get(pk=self.file_uploader.pk).name,
            invalid_data['name']
        )
        self.assertEqual(
            bp_a_uploader_before,
            bp_a_uploader_after
        )
        self.assertEqual(
            bp_b_uploader_before,
            bp_b_uploader_after
        )


@override_settings(DDM_SETTINGS={'EMAIL_PERMISSION_CHECK':  r'.*(\.|@)mail\.com$', })
class TestAPIs(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.base_creds = {
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        }
        base_user = User.objects.create_user(**cls.base_creds)
        base_profile = ResearchProfile.objects.create(
            user=base_user
        )

        cls.no_perm_creds = {
            'username': 'non-perm', 'password': '123', 'email': 'non-perm@mail.com'
        }
        no_perm_user = User.objects.create_user(**cls.no_perm_creds)
        no_perm_profile = ResearchProfile.objects.create(
            user=no_perm_user
        )

        cls.project_base = DonationProject.objects.create(
            name='Base Project',
            slug='base',
            owner=base_profile
        )
        cls.project_alt = DonationProject.objects.create(
            name='Alt Project',
            slug='alternative',
            owner=no_perm_profile
        )
        cls.project_secret = DonationProject.objects.create(
            name='Secret Project',
            slug='super-secret',
            owner=base_profile,
            super_secret=True,
            secret_key='secret_key'
        )

        cls.participant_regular = Participant.objects.create(
            project=cls.project_base,
            start_time=timezone.now()
        )
        cls.participant_secret = Participant.objects.create(
            project=cls.project_secret,
            start_time=timezone.now()
        )

        cls.blueprint_regular = DonationBlueprint.objects.create(
            project=cls.project_base,
            name='donation blueprint',
            expected_fields='"a", "b"',
            file_uploader=None
        )
        cls.blueprint_secret = DonationBlueprint.objects.create(
            project=cls.project_secret,
            name='donation blueprint',
            expected_fields='"a", "b"',
            file_uploader=None
        )

        cls.donation_regular = DataDonation.objects.create(
            project=cls.project_base,
            blueprint=cls.blueprint_regular,
            participant=cls.participant_regular,
            time_submitted=timezone.now(),
            consent=True,
            status='{}',
            data=['data1_pA_bpA', 'data2_pA_bpA']
        )
        cls.donation_secret = DataDonation.objects.create(
            project=cls.project_secret,
            blueprint=cls.blueprint_secret,
            participant=cls.participant_secret,
            time_submitted=timezone.now(),
            consent=True,
            status='{}',
            data=['data1_pA_bpB', 'data2_pA_bpB']
        )

    def test_donation_download_view_valid_login_regular_project(self):
        self.client.login(**self.base_creds)
        url = reverse(
            'ddm_datadonation:download_donation',
            args=[self.project_base.url_id, self.participant_regular.external_id]
        )
        get_response = self.client.get(url, follow=True)
        self.assertEqual(get_response.status_code, 200)

        post_response = self.client.post(url, {})
        self.assertEqual(post_response.status_code, 200)

    def test_donation_download_view_valid_login_secret_project(self):
        self.client.login(**self.base_creds)
        url = reverse(
            'ddm_datadonation:download_donation',
            args=[self.project_secret.url_id, self.participant_secret.external_id]
        )
        get_response = self.client.get(url, follow=True)
        self.assertEqual(get_response.status_code, 200)

        post_response = self.client.post(url, {'secret': 'secret_key'})
        self.assertEqual(post_response.status_code, 200)

        post_response = self.client.post(url, {'secret': 'nonsense'})
        self.assertEqual(post_response.status_code, 422)

    def test_donation_download_view_invalid_login_regular_project(self):
        self.client.login(**self.no_perm_creds)
        url = reverse(
            'ddm_datadonation:download_donation',
            args=[self.project_base.url_id, self.participant_regular.external_id]
        )
        get_response = self.client.get(url)
        self.assertEqual(get_response.status_code, 404)

        post_response = self.client.post(url, {})
        self.assertEqual(post_response.status_code, 404)

    def test_donation_download_view_invalid_login_secret_project(self):
        self.client.login(**self.no_perm_creds)
        url = reverse(
            'ddm_datadonation:download_donation',
            args=[self.project_secret.url_id, self.participant_secret.external_id]
        )
        get_response = self.client.get(url)
        self.assertEqual(get_response.status_code, 404)

        post_response = self.client.post(url, {'secret': 'secret_key'})
        self.assertEqual(post_response.status_code, 404)

    def test_donation_download_view_not_logged_in_regular_project(self):
        client = Client()
        url = reverse(
            'ddm_datadonation:download_donation',
            args=[self.project_base.url_id, self.participant_regular.external_id]
        )
        get_response = client.get(url)
        self.assertRedirects(get_response, reverse('ddm_login'))

        post_response = self.client.post(url, {})
        self.assertRedirects(post_response, reverse('ddm_login'))

    def test_donation_download_view_not_logged_in_secret_project(self):
        client = Client()
        url = reverse(
            'ddm_datadonation:download_donation',
            args=[self.project_secret.url_id, self.participant_secret.external_id]
        )
        get_response = client.get(url)
        self.assertRedirects(get_response, reverse('ddm_login'))

        post_response = self.client.post(url, {'secret': 'secret_key'})
        self.assertRedirects(post_response, reverse('ddm_login'))



# Generated by Django 4.2.16 on 2024-12-04 15:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm_datadonation', '0004_alter_donationinstruction_text'),
    ]

    operations = [
        migrations.AlterField(
            model_name='donationblueprint',
            name='csv_delimiter',
            field=models.CharField(blank=True, default='', help_text='This field allows you to specify the character that separates values in the expected CSV file (e.g., , ; or \\t). If left empty, DDM will try to infer the delimiter from the file structure.', max_length=10),
        ),
        migrations.AlterField(
            model_name='donationblueprint',
            name='expected_fields_regex_matching',
            field=models.BooleanField(default=False, help_text='Select if you use regex expressions in the "Expected fields".'),
        ),
    ]



from copy import deepcopy


ddm_ckeditor =  {
    'blockToolbar': [
        'paragraph', 'heading1', 'heading2', 'heading3',
        '|',
        'bulletedList', 'numberedList',
        '|',
        'blockQuote',
    ],
    'toolbar': [
        'heading', '|',
        'alignment', 'outdent', 'indent', '|',
        'bold', 'italic', 'underline', 'link', 'highlight', '|',
        {
            'label': 'Fonts',
            'icon': 'text',
            'items': ['fontSize', 'fontFamily', 'fontColor']
        }, '|',
        'bulletedList', 'numberedList', 'insertTable', 'blockQuote', 'code', 'removeFormat', '|',
        'insertImage', 'fileUpload', 'mediaEmbed', '|',
        'sourceEditing'
    ],
    'image': {
        'toolbar': ['imageTextAlternative', '|', 'imageStyle:alignLeft',
                    'imageStyle:alignRight', 'imageStyle:alignCenter', 'imageStyle:side',  '|'],
        'styles': [
            'full',
            'side',
            'alignLeft',
            'alignRight',
            'alignCenter',
        ]
    },
    'table': {
        'contentToolbar': [ 'tableColumn', 'tableRow', 'mergeTableCells',
                            'tableProperties', 'tableCellProperties' ],
    },
    'heading' : {
        'options': [
            { 'model': 'paragraph', 'title': 'Paragraph', 'class': 'ck-heading_paragraph' },
            { 'model': 'heading1', 'view': 'h1', 'title': 'Heading 1', 'class': 'ck-heading_heading1' },
            { 'model': 'heading2', 'view': 'h2', 'title': 'Heading 2', 'class': 'ck-heading_heading2' },
            { 'model': 'heading3', 'view': 'h3', 'title': 'Heading 3', 'class': 'ck-heading_heading3' }
        ]
    },
    'htmlSupport': {
            'allow': [
                {
                    'name': 'video',
                    'attributes': {
                        'height': True,
                        'width': True,
                        'controls': True,
                    },
                    'styles': True
                }
            ],
            'disallow': []
        },
    'wordCount': {
        'displayCharacters': False,
        'displayWords': False,
    }
}

ddm_ckeditor_temp_func = deepcopy(ddm_ckeditor)
ddm_ckeditor_temp_func['heading']['options'].append(
    {'model': 'template-func', 'view': 'template-func', 'title': 'Template Functionality', 'class': 'template-func'})



from django.contrib.messages.views import SuccessMessageMixin
from django.urls import reverse_lazy
from django.views.generic.detail import DetailView
from django.views.generic.edit import CreateView, UpdateView, DeleteView
from django.views.generic.list import ListView

from ddm.projects.forms import ProjectCreateForm, ProjectEditForm, BriefingEditForm, DebriefingEditForm
from ddm.projects.models import DonationProject, ResearchProfile
from ddm.auth.views import DDMAuthMixin


class ProjectList(DDMAuthMixin, ListView):
    """ View to display a list of existing donation projects. """
    model = DonationProject
    template_name = 'ddm_projects/list.html'

    def get_queryset(self):
        return DonationProject.objects.filter(owner__user=self.request.user)


class ProjectCreate(SuccessMessageMixin, DDMAuthMixin, CreateView):
    """ View to create a new donation project. """
    model = DonationProject
    template_name = 'ddm_projects/create.html'
    form_class = ProjectCreateForm
    success_message = 'Project was created successfully.'

    def get_initial(self):
        self.initial = super().get_initial()
        self.initial.update({'owner': ResearchProfile.objects.get(user=self.request.user)})
        return self.initial

    def form_valid(self, form):
        form.instance.owner = ResearchProfile.objects.get(user=self.request.user)
        return super().form_valid(form)


class ProjectDetail(DDMAuthMixin, DetailView):
    """ View to display landing page for project. """
    model = DonationProject
    slug_url_kwarg = 'project_url_id'
    slug_field = 'url_id'
    template_name = 'ddm_projects/detail.html'


class ProjectEdit(SuccessMessageMixin, DDMAuthMixin, UpdateView):
    """ View to edit the details of an existing donation project. """
    model = DonationProject
    slug_url_kwarg = 'project_url_id'
    slug_field = 'url_id'
    template_name = 'ddm_projects/edit.html'
    form_class = ProjectEditForm
    success_message = 'Project details successfully updated.'


class ProjectDelete(SuccessMessageMixin, DDMAuthMixin, DeleteView):
    """ View to display a list of existing donation projects. """
    model = DonationProject
    slug_url_kwarg = 'project_url_id'
    slug_field = 'url_id'
    template_name = 'ddm_projects/delete.html'
    success_url = reverse_lazy('ddm_projects:list')
    success_message = 'Project "%s" was deleted.'

    def get_success_message(self, cleaned_data):
        return self.success_message % self.object.name


class BriefingEdit(SuccessMessageMixin, DDMAuthMixin, UpdateView):
    """ View to edit the briefing page. """
    model = DonationProject
    slug_url_kwarg = 'project_url_id'
    slug_field = 'url_id'
    template_name = 'ddm_projects/edit-briefing.html'
    form_class = BriefingEditForm
    success_message = 'Briefing page successfully updated.'


class DebriefingEdit(SuccessMessageMixin, DDMAuthMixin, UpdateView):
    """ View to edit the debriefing page. """
    model = DonationProject
    slug_url_kwarg = 'project_url_id'
    slug_field = 'url_id'
    template_name = 'ddm_projects/edit-debriefing.html'
    form_class = DebriefingEditForm
    success_message = 'Debriefing page successfully updated.'



from django.contrib import admin
from django.urls import reverse
from django.utils.html import format_html
from ddm.projects.models import DonationProject


class DonationProjectAdmin(admin.ModelAdmin):
    """
    Provides an overview of all active Donation Projects.
    """
    list_display = ['name', 'owner', 'date_created', 'edit_link']
    readonly_fields = ['date_created']

    def has_add_permission(self, request, obj=None):
        return False

    @admin.display(description="Link to Project")
    def edit_link(self, obj):
        url = reverse('ddm_projects:detail', args=[obj.pk])
        return format_html(f'<a href="{url}">Show Detail Page</a>')


admin.site.register(DonationProject, DonationProjectAdmin)



# Generated by Django 4.2.16 on 2024-11-07 10:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm_projects', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='donationproject',
            name='briefing_text',
            field=models.TextField(default='', help_text='This text will be displayed to participants in the first step before beginning the data donation. Here, you should introduce your study, who is responsible for the study, provide a quick summary of what your participants will be asked to do in the next steps, etc.', verbose_name='Briefing Text'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='contact_information',
            field=models.TextField(default='', help_text='Please provide the contact information of the person responsible for the conduction of this study (Name, professional address, e-mail, tel, ...). The contact information will be accessible for participants during the data donation. The field is mandatory.', verbose_name='Contact information'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='data_protection_statement',
            field=models.TextField(default='', help_text='Please provide a data protection statement for your data donation collection. This should include the purpose for which the data is collected, how it will be stored, and who will have access to the data. The field is mandatory.', verbose_name='Data Protection Statement'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='debriefing_text',
            field=models.TextField(default='', help_text='This text will be displayed to participants on the last page of the study (i.e., after the data donation and the questionnaire).', verbose_name='Debriefing Text'),
            preserve_default=False,
        ),
    ]



from django.contrib import auth
from django.test import TestCase, override_settings, Client
from django.contrib.auth import get_user_model

from ddm.auth.utils import email_is_valid, user_has_project_access, user_is_permitted
from ddm.projects.models import ResearchProfile, DonationProject

User = get_user_model()


@override_settings(DDM_SETTINGS={'EMAIL_PERMISSION_CHECK':  r'.*(\.|@)mail\.com$', })
class TestAUthUtils(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.valid_creds = {
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        }
        cls.valid_user = User.objects.create_user(**cls.valid_creds)
        cls.valid_profile = ResearchProfile.objects.create(user=cls.valid_user)

        cls.non_permission_creds = {
            'username': 'no_per', 'password': '123', 'email': 'noperm@liam.com'
        }
        cls.user_wo_permission = User.objects.create_user(**cls.non_permission_creds)
        cls.profile_wo_permission = ResearchProfile.objects.create(user=cls.user_wo_permission)

        cls.wo_profile_creds = {
            'username': 'no_prof', 'password': '123', 'email': 'noprof@mail.com'
        }
        cls.user_wo_profile = User.objects.create_user(**cls.wo_profile_creds)

        cls.superuser = User.objects.create_superuser('user', 'some@mail.com', 'password')

    def test_email_is_valid(self):
        self.assertTrue(email_is_valid('some-address@mail.com'))

    def test_email_is_invalid(self):
        self.assertFalse(email_is_valid('some-address@mail.ch'))
        self.assertFalse(email_is_valid('some-address@liam.com'))

    def test_user_has_project_access(self):
        project = DonationProject.objects.create(
            name='test-project', slug='test', owner=self.valid_profile)
        self.assertTrue(user_has_project_access(self.valid_user, project))

    def test_user_has_no_project_access(self):
        project = DonationProject.objects.create(
            name='test-project', slug='test', owner=self.valid_profile)
        self.assertFalse(user_has_project_access(self.user_wo_profile, project))
        self.assertFalse(user_has_project_access(self.user_wo_permission, project))
        anonymous_user = auth.get_user(self.client)
        self.assertFalse(user_has_project_access(anonymous_user, project))

    def test_user_is_permitted(self):
        self.assertTrue(user_is_permitted(self.valid_user))
        self.assertTrue(user_is_permitted(self.superuser))
        self.profile_wo_permission.ignore_email_restriction = True
        self.profile_wo_permission.save()
        self.assertTrue(user_is_permitted(self.user_wo_permission))
        self.assertTrue(user_is_permitted(self.user_wo_profile))

    def test_user_is_not_permitted(self):
        self.assertFalse(user_is_permitted(self.user_wo_permission))
        client = Client()
        anonymous_user = auth.get_user(client)
        self.assertFalse(user_is_permitted(anonymous_user))



# Generated by Django 3.2.13 on 2024-10-25 15:49

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('ddm_participation', '0001_initial'),
        ('ddm_logging', '0002_alter_exceptionlogentry_blueprint'),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.AlterField(
                    model_name='exceptionlogentry',
                    name='participant',
                    field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL,
                                            to='ddm_participation.participant'),
                ),
            ],
            # Reusing an existing table, so do nothing.
            database_operations=[]
        )
    ]



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class DDMCoreConfig(AppConfig):
    default_auto_field = 'django.db.models.AutoField'
    name = 'ddm.core'
    label = 'ddm_core'
    verbose_name = _('DDM Core')

    def ready(self):
        from ddm.datadonation import signals

        # Add default settings for DDM if they are not defined in main application's settings.py.
        from django.conf import settings
        from ddm.core import ckeditor_config

        if not hasattr(settings, 'CKEDITOR_5_CONFIGS'):
            setattr(settings, 'CKEDITOR_5_CONFIGS', {})
        if 'ddm_ckeditor' not in settings.CKEDITOR_5_CONFIGS:
            settings.CKEDITOR_5_CONFIGS['ddm_ckeditor'] = ckeditor_config.ddm_ckeditor
        if 'ddm_ckeditor_temp_func' not in settings.CKEDITOR_5_CONFIGS:
            settings.CKEDITOR_5_CONFIGS['ddm_ckeditor_temp_func'] = ckeditor_config.ddm_ckeditor_temp_func



from django.template.exceptions import TemplateSyntaxError
from django.test import TestCase

from ddm.core.utils.user_content.template import preprocess_user_content, render_user_content


class TestPreprocessUserContent(TestCase):
    def test_replacement(self):
        content = ''
        self.assertEqual(preprocess_user_content(content), '')

        content = '&gt;'
        self.assertEqual(preprocess_user_content(content), '>')

        content = '&lt;'
        self.assertEqual(preprocess_user_content(content), '<')

        content = 'Some text and random&lt;symbol and random&gt;-symbol.'
        self.assertEqual(preprocess_user_content(content), 'Some text and random<symbol and random>-symbol.')


class TestRenderUserContent(TestCase):
    """
    Test ensures that unsafe tags and filters are not added by mistake and that the most important allowed tags and
    filters are available.
    """
    def test_include_tag_is_invalid(self):
        with self.assertRaises(TemplateSyntaxError):
            render_user_content('{% include "something.html" %}')

    def test_load_tag_is_invalid(self):
        with self.assertRaises(TemplateSyntaxError):
            render_user_content('{% load static %}')

    def test_debug_tag_is_invalid(self):
        with self.assertRaises(TemplateSyntaxError):
            render_user_content('{% debug %}')

    def test_csrf_token_tag_is_invalid(self):
        with self.assertRaises(TemplateSyntaxError):
            render_user_content('{% csrf_token %}')

    def test_extends_tag_is_invalid(self):
        with self.assertRaises(TemplateSyntaxError):
            render_user_content('{% extends "something.html" %}')

    def test_autoescape_tag_is_invalid(self):
        with self.assertRaises(TemplateSyntaxError):
            render_user_content('{% autoescape off %}')

    def test_url_tag_is_invalid(self):
        with self.assertRaises(TemplateSyntaxError):
            render_user_content('{% url "some url name" %}')

    def test_safe_filter_is_invalid(self):
        with self.assertRaises(TemplateSyntaxError):
            render_user_content('{{ "<script>function someFun() {}</script>"|safe }}')

    def test_safeseq_filter_is_invalid(self):
        with self.assertRaises(TemplateSyntaxError):
            context = {'some_list': [1, 2, 3]}
            render_user_content('{{ some_list|safeseq }}', context)

    def test_escapejs_filter_is_invalid(self):
        with self.assertRaises(TemplateSyntaxError):
            render_user_content('{{ "value"|escapejs }}')

    def test_if_tag_is_valid(self):
        self.assertEqual('first', render_user_content('{% if 1 < 2 %}first{% endif %}'))

    def test_for_tag_is_valid(self):
        context = {'some_list': [1, 2, 3]}
        self.assertEqual('123', render_user_content('{% for i in some_list %}{{i}}{% endfor %}', context))



from ddm.logging.models import ExceptionLogEntry, ExceptionRaisers


def log_server_exception(project, message):
    ExceptionLogEntry.objects.create(
        project=project,
        raised_by=ExceptionRaisers.SERVER,
        message=message
    )
    return
